/* sinfo -- Structure Info (SInfo) program to retrieve final tables
 * for crystal structure analyses from a relational database. The program
 * may either be run as a traditional unix command line program or as
 * a CGI binary.   If the program is invoked without any command line
 * args then it runs as a CGI binary. If there are any command line
 * args, the program behaves as a traditional command line program.
 *
 * A sequentially executing version may be compiled by specifying
 * CFLAGS -DSERIAL_EXEC
 *
 * Written by:
 *	Paul D. Boyle
 *	Department of Chemistry
 *	Box 8204
 *	North Carolina State University
 *	Raleigh, NC 27695-8204
 *	March 1998
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */
#define _XOPEN_SOURCE

#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <getopt.h>
#include <pwd.h>
#include <sys/stat.h>
#include <sys/wait.h>

#ifndef  SERIAL_EXEC
#include <sys/ipc.h>
#include <sys/msg.h>
#endif
#include <errno.h>

#include "cgi_util.h"

#include "sinfo.h"
#include "qthread.h"
#include "output.h"

#ifndef SERIAL_EXEC
#include "sinfo_ipc.h"
#else
#include "process_tuples.h"
#endif

#define PUTENV_STR  "PGUSER=%s"
#define DOUBLE_QUOTE 0x22


void Usage( char *prog );

int main( int argc, char **argv )
{
   extern int run_mode;
   int i_opt;
   char *pg_hostname = NULL,
        *pg_database = NULL,
        *pg_user = NULL,
        *default_pg_user = NULL;

   size_t  len;
   size_t  content_length;
   time_t  now;            /* place to store current time. */ 
           
   union   cgi_data *entry;    /* defined in cgi_util.h */

   char    *cl_ptr,
            *buffer,
            *ptr;
   char    error_msg[80];            /* for formatting errors with sprintf()
                                       * used in ReportErrorHTML()
                                       */

   int     n_field,
           ret_val;
   char    user[DB_STD1_LEN];        /* to save the user name */
   char    working_dir[256];         /* HARDCODED -- I am bad */
   char    pguser[DB_STD1_LEN],
           this_host[DB_HOST_LEN],   /* to get name of this server */
           structure_id_code[DB_STD2_LEN];

   struct passwd *user_info;

   int     i = 0,
           j = 0;

   Query   queries[N_TABLES];   /* typedef Query in qthread.h */
   enum TableType tab_type;

#ifndef SERIAL_EXEC
/* typedefs for reading the table specific data structs */

    Lattice       lattice_data; 
    UnitCell      unitcell_data;
    Formula       *formula_data = NULL;
    Experimental  exper_data; 
    Refinement    refine_data;

/* structures for the messages */

struct latt_msg  msg_data_latt;
struct cell_msg  msg_data_cell;
struct form_msg  msg_data_form;
struct expr_msg  msg_data_expr;
struct refn_msg  msg_data_refn;


    int     msg_id,  /* identifier for message queue */
            wait_status,
            n_errors,
            ipc_requests_processed;
    pid_t   waited_pid;

    void    *recvd_buffer;      /* for reading
                                 * received messages in queue
                                 */

#else  /* for serial queries of data base */
    Tuple *tup_latt,
          *tup_cell, 
          *tup_form,
          *tup_expr,
          *tup_refn,
          *tup_atom,
          *tup_b_len, 
          *tup_b_ang;

    int n_tuples[N_TABLES],
        n_fields[N_TABLES];
 
    Lattice      *lattice_data;
    UnitCell     *unitcell_data;
    Formula      *formula_data;
    Experimental *exper_data;
    Refinement   *refine_data; 
    AtomCoords   *atoms;
    BondLength   *bonds;
    BondAngle    *angles;
#endif

    int           n_atom_types = 0; /* obtained from the Formula 
                                     * query child process.
                                     */

/* first establish run mode.  If argc != then we are invoked from
 * the command line.  If argc == 1 then we assume we are running as
 * a CGI binary.
 */
   if( argc != 1 ) 
       run_mode = CLI_MODE;
   else
       run_mode = CGI_MODE;

   if( (run_mode == CGI_MODE) && (NULL == getenv("CONTENT_LENGTH")) ) {
        Usage( argv[0] );
        fputs( "Exiting.\n", stderr );
        exit( EXIT_FAILURE );
   }

/* for CLI_MODE structure id code should be
 * last argument on the command line
 */

   if( CLI_MODE == run_mode ) {
       strcpy( structure_id_code, argv[argc-1] );
    
    /* get current user to be the default pg_user */
       user_info = getpwuid( getuid() );
       len = strlen( user_info->pw_name );
       default_pg_user = malloc( len + 1 );
       if( !default_pg_user ) {
         fprintf( stderr, "Memory Allocation Error: %s:%d\n", __FILE__, 
                          __LINE__ );
         exit( EXIT_FAILURE );
       }
       strcpy( default_pg_user, user_info->pw_name );
       pg_user = default_pg_user;
    
    /* process the command line options */
        while( (i_opt = getopt( argc, argv,"h:d:u:")) != EOF ) {
          switch( i_opt ) {
             case 'h':    /* hostname on which database resides */
                len = strlen( optarg );
                pg_hostname = malloc( len + 1 );
                if( !pg_hostname ) {
                  fprintf( stderr, "malloc() error in %s:%d. Exiting\n",
                           __FILE__, __LINE__ );
                  exit( EXIT_FAILURE );
                }
                strcpy( pg_hostname, optarg );
                pg_hostname[len] = '\0';
                break;
             case 'd':    /* name of database */
                len = strlen( optarg );
                pg_database = malloc( len + 1 );
                if( !pg_database ) {
                  fprintf( stderr, "malloc() error in %s:%d. Exiting\n",
                           __FILE__, __LINE__ );
                  exit( EXIT_FAILURE );
                }
                strcpy( pg_database, optarg );
                pg_database[len] = '\0';
                break;
             case 'u':    /* user of data base */
                len = strlen( optarg );
                pg_user = malloc( len + 1 );
                if( !pg_user ) {
                  fprintf( stderr, "malloc() error in %s:%d. Exiting\n",
                           __FILE__, __LINE__ );
                  exit( EXIT_FAILURE );
                }
                strcpy( pg_user, optarg );
                pg_user[len] = '\0';
                free( default_pg_user );  /* don't need this anymore */
                break;
              default:
                fprintf( stderr, "Error: -%c -- unrecognized option.\n", i );
                break;
          }
        }
    
    /* set the PostgreSQL user name for database authorization and
     * fill the db_host, db_name structure members
     * for the queries.
     */
        sprintf( pguser, PUTENV_STR, pg_user );
        putenv( pguser );
    
        for( j = LATT; j < N_TABLES; j++ ) {
           strcpy( queries[j].db_host, pg_hostname );
           strcpy( queries[j].db_name, pg_database );
        }
        free( pg_hostname );
        free( pg_database );
    }
    else {  /* running as CGI bin */
    /* first get the content length from the environment variable */
    
        cl_ptr = getenv( "CONTENT_LENGTH" );
        if( !cl_ptr ) {
            ReportErrorHTML( "CONTENT_LENGTH variable not set." );
            exit( EXIT_SUCCESS );
        }
        content_length = (size_t)atoi( cl_ptr );
    
    /* check for errors with METHOD and CONTENT_TYPE */
    
        switch( VerifyForm("POST") ) {
          case 0:   /* everything ok */
             break;
          case 1:
             ReportErrorHTML( "REQUEST_METHOD incorrectly set." );
             exit( EXIT_SUCCESS );
          case 2:
             ReportErrorHTML( "CONTENT_TYPE incorrectly set." );
             exit( EXIT_SUCCESS );
          case 3:
             ReportErrorHTML( "REQUEST_METHOD and CONTENT_TYPE incorrectly set." );
             exit( EXIT_SUCCESS );
          default:  /* should never reach here */
             ReportErrorHTML( "Unrecognized Error Condition. Exiting" );
             exit( EXIT_SUCCESS );
        }
    
    /* allocate memory for read in buffer and fill it with data */
        buffer = malloc( content_length + 1 );
        if( !buffer ) {
            ReportErrorHTML( "Read buffer memory allocation error." );
            exit( EXIT_FAILURE );
        }
        buffer[content_length] = '\0';
        fread( buffer, content_length, 1, stdin );
    
    
    /* get number of fields and dynamically allocate an array of unions for
     * the data.
     */
        n_field = CountFields( buffer );
    
        entry = malloc( n_field * sizeof(union cgi_data) );
        if( !entry ) {
            ReportErrorHTML( "Memory allocation error" );
            exit( EXIT_FAILURE );
        }
    
    /* chop the buffer into data fields and process the name value pairs */
        ptr = buffer;
        i = 0;
        for( ptr = strtok(buffer, FIELD_TOKEN_STR);
            NULL != ptr;
            ptr = strtok(NULL, FIELD_TOKEN_STR )) {
              entry[i].field = malloc( strlen(ptr) + 1 );
    
             if( !entry[i].field ) {
                 ReportErrorHTML( "Memory Allocation Error. Exiting." );
                 exit( EXIT_FAILURE );
             }
             strcpy( entry[i].field, ptr);
             PlusToSpace( entry[i].field );
             entry[i].field = TranslateEscapes( entry[i].field );
             entry[i].field = TranslateNewlines( entry[i].field );
             entry[i].pair.name = entry[i].field;
             entry[i].pair.value = ParseNameValuePairs( entry[i].field );
             i++;
        }
    
    /* Get the user name for PostgreSQL and name of the database and server. 
     * Note that the first argument of the strcmp()'s must match the "values"
     * which are listed in the HTML form being processed.
     */
    
       for( i = 0; i < n_field; i++ ) {
          if( 0 == strcmp(SERVER, entry[i].pair.name) ) {
            if( DB_HOST_LEN < strlen( entry[i].pair.value ) ) /* overflow test */
                BufferOverFlowError( entry, n_field );
    
            for( j = LATT; j < N_TABLES; j++ ) 
                 strcpy( queries[j].db_host, entry[i].pair.value );
          }
          if( 0 == strcmp(DB_NAME, entry[i].pair.name) ) {
            if( DB_STD1_LEN < strlen( entry[i].pair.value ) ) /* overflow test */
                BufferOverFlowError( entry, n_field );
    
            for( j = LATT; j < N_TABLES; j++ ) 
                 strcpy( queries[j].db_name, entry[i].pair.value );
            
           /* save the name of the user for creating the appropriate
            * working directory.
            */
            strcpy( user, entry[i].pair.value );
          }
          if( 0 == strcmp(USER_NAME, entry[i].pair.name) ) {
            if( DB_STD1_LEN < (strlen(entry[i].pair.value) +
                               strlen(PUTENV_STR) ) ) /* overflow test */
                BufferOverFlowError( entry, n_field );
    
              sprintf( pguser, PUTENV_STR, entry[i].pair.value );
              putenv( pguser );
    
          }
          if( 0 == strcmp(STRUCT_NAME, entry[i].pair.name) ) {
            if( DB_STD2_LEN < strlen( entry[i].pair.value ) ) /* overflow test */
                BufferOverFlowError( entry, n_field );
    
              strcpy( structure_id_code, entry[i].pair.value );
          }
       }
    /* change the current working directory to the appropriate place, based
     * on user name. We need to get the home directory of the web server,
     * and a relative path to where the pages will be deposited.
     */
    
       user_info = getpwuid( getuid() );
       sprintf( working_dir, "%s/pages/tables/%s/%s/", user_info->pw_dir, user,
                             structure_id_code );
       ret_val = mkdir( working_dir, S_IRUSR | S_IWUSR | S_IXUSR );
       if( -1 == ret_val ) {
         if( EEXIST != errno ) {
           sprintf( error_msg, "%s: %s. Exiting\n", working_dir, strerror(errno) );
           ReportErrorHTML( error_msg );
           exit( EXIT_FAILURE );
         }
       }
       chdir( working_dir );
   }
                         

#ifndef SERIAL_EXEC
/* Create message queue for process and it's subsequent children */
   msg_id = msgget( IPC_PRIVATE, IPC_CREAT | PERMS ); 
   if( -1 == msg_id ) {
     switch(run_mode) {
       case CLI_MODE:
          perror( "msgget()" );
          break;
       case CGI_MODE:
          sprintf( error_msg, "msgget():  %s\n", strerror(errno) );
          ReportErrorHTML( error_msg );
          break;
     }
     ret_val = rmdir( working_dir );
     exit( EXIT_FAILURE );
   }
    
/* initialize the SQL buffers.  Fork off the children and let them
 * so their thing:  bond_length, bond_angle, atom_coords are
 * "fire and forget", while the lattice, unit_cell, formula, 
 * experimental, and refinement report back to parent the
 * results of their searches via the shared memory mechanism 
 * initialized above.
 */
    for( tab_type = LATT; tab_type < N_TABLES; tab_type++ ) {
         Query *query = &queries[tab_type];
         pid_t pid;
         char *table_names[] = { "lattice", "unit_cell",  "formula",
                                 "experimental", "refinement", "atom_coords",
                                 "bond_length", "bond_angle"};

         if( tab_type != UNIT_CELL ) {
            query->query_type = tab_type;
            sprintf(  query->query_buffer, WHOLE_TABL,
                           table_names[tab_type], table_names[tab_type],
                           table_names[tab_type], structure_id_code );
         }
         else {
            query->query_type = tab_type;
            sprintf( query->query_buffer, "SELECT unit_cell.* FROM unit_cell "
                                          "WHERE unit_cell.id = '%s';",
                                          structure_id_code );
         }

         pid = fork();
         switch( pid ) {
           case -1:
              if( CLI_MODE == run_mode )
                  perror( "fork()" );
              else {
                 sprintf( error_msg, "Child searching %s table failed to fork(). Exiting.\n",
                                      table_names[tab_type] );
                 ReportErrorHTML( error_msg );
              }
              ret_val = rmdir( working_dir );
              exit( EXIT_FAILURE );
              break;
           case 0:  /* child does this */
              query->msg_id = msg_id;
              WholeTableThread( query );
              break;
          default:
#ifdef DEBUG
              query->pid = pid;
              fprintf( stdout, "Child %d searching %s table.\n",
                       query->pid, table_names[tab_type] );
#endif            
              break;
         }
    }

/* Monitor which  message queues have been filled by 
 * children and are ready for reading.  Collect the data.
 */

    ipc_requests_processed = 0;
    n_errors = 0;
    recvd_buffer = malloc( QUEUE_SIZ );
    if( !recvd_buffer ) {
     if( CLI_MODE == run_mode )  
         fprintf( stderr, "Memory Allocation Error, %s:%d\n",
                          __FILE__, __LINE__ );
     else 
         ReportErrorHTML( "Memory Allocation Error.\n" );

     ret_val = msgctl( msg_id, IPC_RMID, NULL );
     if( -1 == ret_val ) {  
       if( CLI_MODE == run_mode ) {
          perror( "msgctl()" );
          fprintf( stderr, "Remove message queue manually.\n" );
       }
       else {
          sprintf( error_msg, "msgctl(): %s\n Web admin: remove queue by hand.\n",
                   strerror(errno) );
          ReportErrorHTML( error_msg );
       }
         
     }
     ret_val = rmdir( working_dir );
     exit( EXIT_FAILURE );
    }    

    while( N_MESG_TYPES - ipc_requests_processed  ) {
        long code = 0L;
        ret_val = msgrcv( msg_id, recvd_buffer, QUEUE_SIZ, -ERR_MSG, 0 );
        memcpy( &code, recvd_buffer, sizeof(long) );
        switch( code ) {
          case LATT_MSG:
             msg_data_latt = *(struct latt_msg *)recvd_buffer;
             lattice_data = msg_data_latt.latt_data;
             ipc_requests_processed++;
             break;
          case CELL_MSG:
             msg_data_cell = *(struct cell_msg *)recvd_buffer;
             unitcell_data = msg_data_cell.cell_data;
             ipc_requests_processed++;
             break;
          case FORM_MSG:
             msg_data_form = *(struct form_msg *)recvd_buffer;
             n_atom_types = msg_data_form.n_bytes / sizeof( Formula );
             formula_data = malloc( n_atom_types * sizeof(Formula) );
             if( !formula_data ) {
                switch( run_mode ) { 
                  case CLI_MODE:
                     fprintf( stderr, "Memory Allocation Error, %s:%d\n",
                                      __FILE__, __LINE__ );
                     break;
                     case CGI_MODE:
                        ReportErrorHTML( "Memory Allocation Error.\n" );
                        break;
                 }
                 exit( EXIT_FAILURE );
             }
             memcpy( formula_data, msg_data_form.formula_data,
                                   (unsigned)msg_data_form.n_bytes );    
             ipc_requests_processed++;
             break;
          case EXPR_MSG:
             msg_data_expr = *(struct expr_msg *)recvd_buffer;
             exper_data = msg_data_expr.exper_data;
             ipc_requests_processed++;
             break;
          case REFN_MSG:
             msg_data_refn = *(struct refn_msg *)recvd_buffer;
             refine_data = msg_data_refn.refine_data;
             ipc_requests_processed++;
             break;
          case ERR_MSG:  /* parent removes message queue and exits */
             ipc_requests_processed++;
             n_errors++;
             waited_pid = waitpid( -1, &wait_status, 0 );
             if( CLI_MODE == run_mode ) {
                 if( WIFEXITED(wait_status) ) {
                   switch( WEXITSTATUS(wait_status) ) {
                     case LATT_MSG:
                        fprintf( stderr, "Child %d (lattice search) failed.\n",
                                 waited_pid );
                        break;
                     case CELL_MSG:
                        fprintf( stderr, "Child %d (cell search) failed.\n",
                                 waited_pid );
                        break;
                     case FORM_MSG:
                        fprintf( stderr, "Child %d (formula search) failed.\n",
                                 waited_pid );
                        break;
                     case EXPR_MSG:
                        fprintf( stderr, "Child %d (experimental search) failed.\n",
                                 waited_pid );
                        break;
                     case REFN_MSG:
                        fprintf( stderr, "Child %d (refinement search) failed.\n",
                                 waited_pid );
                        break;
                   }
                 }
             }
             break;
        }

    }          /* end of while() loop */
 

    if( n_errors ) {  /* there were problem children, clean up and bail out */
      ret_val = msgctl( msg_id, IPC_RMID, NULL );
      if( -1 == ret_val ) { 
          switch(run_mode) {
            case CLI_MODE:
               perror( "msgctl()" );
               fprintf( stderr, "Remove message queue manually.\n" );
               break;
            case CGI_MODE:
               sprintf( error_msg, "msgctl(): %s\n Web admin: remove queue by hand.\n",
                                   strerror(errno) );
               ReportErrorHTML( error_msg );
               break;
            }
      }    
      switch( run_mode ) {
        case CLI_MODE:
           fprintf( stderr, "Jobs encountered %d errors.  Processing terminated.\n", 
                            n_errors );
           break;
        case CGI_MODE:
           sprintf( error_msg, "Errors occurred during search.  Terminated.\n" );
           ReportErrorHTML( error_msg );
           break;
      }
      ret_val = rmdir( working_dir );
      exit( EXIT_FAILURE );
    }


/* Parent removes message queue */
         ret_val = msgctl( msg_id, IPC_RMID, NULL );
         if( -1 == ret_val ) { 
             switch(run_mode) {
               case CLI_MODE:
                  perror( "msgctl()" );
                  fprintf( stderr, "Remove message queue manually.\n" );
                  break;
               case CGI_MODE:
                  sprintf( error_msg, "msgctl(): %s\n Web admin: remove queue by hand.\n",
                                      strerror(errno) );
                  ReportErrorHTML( error_msg );
                  break;
             }
         }    
         
         

#else  /* we execute queries sequentially */
/* initialize the SQL buffers, and do queries */  

    for( tab_type = LATT; tab_type < N_TABLES; tab_type++ ) {

         char *table_names[] = { "lattice", "unit_cell",  "formula",
                                 "experimental", "refinement", "atom_coords",
                                 "bond_length", "bond_angle"};

         if( tab_type != UNIT_CELL ) {
            queries[tab_type].query_type = tab_type;
            sprintf(  queries[tab_type].query_buffer, WHOLE_TABL,
                           table_names[tab_type], table_names[tab_type],
                           table_names[tab_type], structure_id_code );
         }
         else {
            queries[tab_type].query_type = tab_type;
            sprintf( queries[tab_type].query_buffer, "SELECT unit_cell.* FROM unit_cell "
                                          "WHERE unit_cell.id = '%s';",
                                          structure_id_code );
         }

         switch( tab_type ) {
           case LATT:
              tup_latt = SearchDatabase( &queries[tab_type], &n_tuples[tab_type],
                                         &n_fields[tab_type]);
              if( !tup_latt ) {
                 if( CLI_MODE == run_mode ) 
                     fputs( "Query unsuccessful. Exiting.\n", stderr );
                 else 
                     ReportErrorHTML( "Query unsuccessful. Exiting.\n" );

                 exit( EXIT_FAILURE );
              }

              lattice_data = ProcessLatticeTuple( tup_latt, n_tuples[tab_type],
                                                  n_fields[tab_type]);
              break;
           case UNIT_CELL:
              tup_cell = SearchDatabase( &queries[tab_type], &n_tuples[tab_type],
                                         &n_fields[tab_type]);
              if( !tup_cell ) {
                 if( CLI_MODE == run_mode ) 
                     fputs( "Query unsuccessful. Exiting.\n", stderr );
                 else 
                     ReportErrorHTML( "Query unsuccessful. Exiting.\n" );

                  exit( EXIT_FAILURE );
              }
              unitcell_data = ProcessUnitCellTuple( tup_cell, n_tuples[tab_type],
                                                  n_fields[tab_type]);
              break;
           case FORM:
              tup_form = SearchDatabase( &queries[tab_type], &n_tuples[tab_type],
                                         &n_fields[tab_type]);
              if( !tup_form ) {
                 if( CLI_MODE == run_mode ) 
                     fputs( "Query unsuccessful. Exiting.\n", stderr );
                 else 
                     ReportErrorHTML( "Query unsuccessful. Exiting.\n" );

                  exit( EXIT_FAILURE );
              }
              formula_data = ProcessFormulaTuple( tup_form, n_tuples[tab_type],
                                                  n_fields[tab_type]);
              n_atom_types = n_tuples[tab_type];
              break;
           case EXPER:
              tup_expr = SearchDatabase( &queries[tab_type], &n_tuples[tab_type],
                                         &n_fields[tab_type]);
              if( !tup_expr ) {
                 if( CLI_MODE == run_mode ) 
                     fputs( "Query unsuccessful. Exiting.\n", stderr );
                 else 
                     ReportErrorHTML( "Query unsuccessful. Exiting.\n" );

                  exit( EXIT_FAILURE );
              }
              exper_data = ProcessExperimentalTuple( tup_expr, n_tuples[tab_type],
                                                  n_fields[tab_type]);
              break;
           case REFINE:
              tup_refn = SearchDatabase( &queries[tab_type], &n_tuples[tab_type],
                                         &n_fields[tab_type]);
              if( !tup_refn ) {
                 if( CLI_MODE == run_mode ) 
                     fputs( "Query unsuccessful. Exiting.\n", stderr );
                 else 
                     ReportErrorHTML( "Query unsuccessful. Exiting.\n" );

                  exit( EXIT_FAILURE );
              }
              refine_data = ProcessRefinementTuple( tup_refn, n_tuples[tab_type],
                                                  n_fields[tab_type]);
              break;
           case ATOM_COORDS:
              tup_atom = SearchDatabase( &queries[tab_type], &n_tuples[tab_type],
                                         &n_fields[tab_type]);
              if( !tup_atom ) {
                 if( CLI_MODE == run_mode ) 
                     fputs( "Query unsuccessful. Exiting.\n", stderr );
                 else 
                     ReportErrorHTML( "Query unsuccessful. Exiting.\n" );

                  exit( EXIT_FAILURE );
              }
              atoms = ProcessAtomCoordsTuple( tup_atom, n_tuples[tab_type],
                                                  n_fields[tab_type]);
              ProduceAtomParamsFile( structure_id_code,
                                     atoms, unitcell_data, n_tuples[tab_type] );
              break;
           case BOND_LENGTH:
              tup_b_len = SearchDatabase( &queries[tab_type], &n_tuples[tab_type],
                                         &n_fields[tab_type]);
              if( !tup_b_len ) {
                 if( CLI_MODE == run_mode ) 
                     fputs( "Query unsuccessful. Exiting.\n", stderr );
                 else 
                     ReportErrorHTML( "Query unsuccessful. Exiting.\n" );

                  exit( EXIT_FAILURE );
              }
              bonds = ProcessBondLengthTuple( tup_b_len, n_tuples[tab_type],
                                                  n_fields[tab_type]);
              MolecularGeometryTable( structure_id_code, bonds, n_tuples[tab_type], BOND_LENGTH );
              break;
           case BOND_ANGLE:
              tup_b_ang = SearchDatabase( &queries[tab_type], &n_tuples[tab_type],
                                         &n_fields[tab_type]);
              if( !tup_b_ang ) {
                 if( CLI_MODE == run_mode ) 
                     fputs( "Query unsuccessful. Exiting.\n", stderr );
                 else 
                     ReportErrorHTML( "Query unsuccessful. Exiting.\n" );

                  exit( EXIT_FAILURE );
              }
              angles = ProcessBondAngleTuple( tup_b_ang, n_tuples[tab_type],
                                                  n_fields[tab_type]);
              MolecularGeometryTable( structure_id_code, angles, n_tuples[tab_type], BOND_ANGLE );
              break;
          default:    /* do nothing */
              break;
         }
    }

#endif  /* EXEC_SERIAL */

/* now all information should be gathered for producing the crystal
 * data summary table.
 */
#ifdef SERIAL_EXEC
    ProduceCrystalDataFile( n_atom_types, structure_id_code, 
                            formula_data, lattice_data,
                            unitcell_data, exper_data,
                            refine_data ); 
#else
    ProduceCrystalDataFile( n_atom_types, structure_id_code, 
                            formula_data, &lattice_data,
                            &unitcell_data, &exper_data,
                            &refine_data ); 
#endif

/* tell where and what the output is */
   switch( run_mode ) {
     case CLI_MODE:   
        fprintf( stdout, "The following output files were produced:\n"
                         "\t"SUMMARY_SUFFIX_TXT"\n"
                         "\t"XYZ_SUFFIX_TXT"\n"
                         "\t"ADP_SUFFIX_TXT"\n"
                         "\t"BOND_SUFFIX_TXT"\n"
                         "\t"ANGLE_SUFFIX_TXT"\n", structure_id_code, structure_id_code,
                           structure_id_code, structure_id_code, structure_id_code );
        break;
                 
     case CGI_MODE: 
        time( &now );
        gethostname( this_host, DB_HOST_LEN );
        StdContentHeader( "text" );
        StdHTMLHeader();
   
        fprintf( stdout, "<H1 ALIGN=CENTER>Final Tables for %s</H1><br>\n<ul>\n",
                          structure_id_code );
     
        fprintf( stdout, "<li><a href=%c/tables/%s/%s/"SUMMARY_SUFFIX_HTML"%c>"
                          "<strong>Crystal Summary Table</strong></a>", DOUBLE_QUOTE, user,
                           structure_id_code, structure_id_code, DOUBLE_QUOTE );
     
        fprintf( stdout, "<li><a href=%c/tables/%s/%s/"XYZ_SUFFIX_HTML"%c>" 
                          "<strong>Atomic Coordinate Table</strong></a>", DOUBLE_QUOTE, user,
                           structure_id_code, structure_id_code, DOUBLE_QUOTE );
     
        fprintf( stdout, "<li><a href=%c/tables/%s/%s/"ADP_SUFFIX_HTML"%c>"
                          "<strong>Anisotropic Displacement Parameter Table</strong></a>",
                           DOUBLE_QUOTE, user,
                           structure_id_code, structure_id_code, DOUBLE_QUOTE );
     
        fprintf( stdout, "<li><a href=%c/tables/%s/%s/"BOND_SUFFIX_HTML"%c>"
                          "<strong>Bond Length Table</strong></a>", DOUBLE_QUOTE, user,
                           structure_id_code, structure_id_code, DOUBLE_QUOTE );
     
        fprintf( stdout, "<li><a href=%c/tables/%s/%s/"ANGLE_SUFFIX_HTML"%c>"
                          "<strong>Bond Angle Table</strong></a>", DOUBLE_QUOTE, user,
                           structure_id_code, structure_id_code, DOUBLE_QUOTE );
        fprintf( stdout, "\n</ul><hr>" );

        fprintf( stdout, "<address>Produced by SINFO<br>at %s<br>by %s</address>\n",
                         ctime(&now), this_host );

        StdHTMLFooter();
        break;
    }

    exit( EXIT_SUCCESS );
}

void Usage( char *prog )
{
    int i;
    static char *help[] = {
           "Options:",
           "    -h <hostname>",
           "    -d <database name>",
           "    -u <user name>",
           "" };
   fprintf( stderr, "Usage: %s [options] structure_id_code\n", prog );
   for( i = 0; *help[i]; i++ ) 
      fprintf( stderr, "%s\n", help[i] );
}
