/* functions to produce formatted output files.
 * Written by:
 *	Paul D. Boyle
 *	Department of Chemistry
 *	Box 8204
 *	North Carolina State University
 *	Raleigh, NC 27695-8204
 *
 *	March 1998
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#define _XOPEN_SOURCE

#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <errno.h>
#include <unistd.h>

#include "sinfo.h"
#include "qthread.h"
#include "output.h"
#include "cgi_util.h"

#define LINE_LEN        78
#define LONG_LINE_LEN  132
#define SPACE         0x20

/* globals */
extern int run_mode;

/* This function produces a "Crystal Data Summary" as commonly seen
 * in chemical publications.
 */
void ProduceCrystalDataFile( int n_atom_types, 
                             char *struct_name,
                             Formula *formula,
                             Lattice *latt,
                             UnitCell *unit,
                             Experimental *exper, 
                             Refinement  *refine )
{
    
    FILE *summary_file;

    char *summary_file_name,
         *summary_title,
         *summary_suffix,
         *chemical_formula,
         *ptr;
         

     char   error_msg[80];
     double formula_weight;
     size_t len1,
            len2;

     enum CrystalFamily crystal_family = TRICLINIC;

   if( CLI_MODE == run_mode ) 
      summary_suffix = SUMMARY_SUFFIX_TXT;
   else
      summary_suffix = SUMMARY_SUFFIX_HTML;
       

/* build up the output file names and open them. */

    summary_file_name = malloc( strlen(summary_suffix) + strlen(struct_name) + 1 );
    if( !summary_file_name ) {
      if( CLI_MODE == run_mode ) 
        fprintf( stderr, "Memory allocation: %s:%d\n", 
                         __FILE__, __LINE__ );
      else  
        ReportErrorHTML( "Memory Allocation" );
      
      exit( EXIT_FAILURE );
    }
    sprintf( summary_file_name, summary_suffix, struct_name );


    summary_file = fopen( summary_file_name, "w" );
    if( !summary_file ) {
       if( CLI_MODE == run_mode ) {
           fprintf( stderr, "fopen(): %s:%d\n", __FILE__, __LINE__ );
           perror( summary_file_name );
       }
       else {
           sprintf( error_msg, "Can't open %s: %s\n", summary_file_name, 
                    strerror(errno) );
           ReportErrorHTML( error_msg );
       } 
       exit( EXIT_FAILURE );
    }


/* Build up the title strings */

    summary_title = malloc( strlen(SUMMARY_TITLE) + strlen(struct_name) + 1 );
    if( !summary_title ) {
      if( CLI_MODE == run_mode ) 
        fprintf( stderr, "Memory allocation: %s:%d\n", 
                         __FILE__, __LINE__ );
      else  
        ReportErrorHTML( "Memory Allocation" );
      
      exit( EXIT_FAILURE );
    }
    sprintf( summary_title, SUMMARY_TITLE, struct_name );



/* Write the standard html headers to the files along with the titles */

    if( CLI_MODE == run_mode ) 
       fprintf( summary_file, "%s\n\n", summary_title );
    else {
       fprintf( summary_file, STD_HTML_HEADER, summary_title, summary_title );
       fprintf( summary_file, "<pre>\n" );
    }

/* Get the formula and formula weight data from the Formula pointer. */

    chemical_formula = FormulaString( formula, n_atom_types );
    formula_weight = CalcFormulaWeight( formula, n_atom_types );

/* calculate the sigma for the unit cell volume */
   (void)CalcSigmaCellVolume( unit );

/* Fix up the cases on some text and get the proper crystal family
 * for the switch() statement.
 */
  
   ptr = latt->crystal_family;
   *ptr = toupper( *ptr );
   ptr++;
   while( *ptr ) {
       *ptr = tolower( *ptr );
       ptr++;
   }

   ptr = latt->space_group;
   *ptr = toupper( *ptr );
   ptr++;
   while( *ptr ) {
       *ptr = tolower( *ptr );
       ptr++;
   }

   if( 0 == strcmp("Triclinic", latt->crystal_family) )
       crystal_family = TRICLINIC;
    else if( 0 == strcmp("Monoclinic", latt->crystal_family) )
       crystal_family = MONOCLINIC;
    else if( 0 == strcmp("Orthorhombic", latt->crystal_family) )
       crystal_family = ORTHORHOMBIC;
    else if( 0 == strcmp("Tetragonal", latt->crystal_family) )
       crystal_family = TETRAGONAL;
    else if( 0 == strcmp("Rhombohedral", latt->crystal_family) )
       crystal_family = RHOMBOHEDRAL;
    else if( 0 == strcmp("Trigonal", latt->crystal_family) )
       crystal_family = TRIGONAL;
    else if( 0 == strcmp("Hexagonal", latt->crystal_family) )
       crystal_family = HEXAGONAL;
    else if( 0 == strcmp("Cubic", latt->crystal_family) )
       crystal_family = CUBIC;

/* begin printing stuff out */
    ptr = "Formula";
    len1 = strlen( ptr );
    len2 = strlen( chemical_formula );
    fprintf( summary_file,"%*s%*c%*s\n", len1, ptr,
                          LINE_LEN - (len1 + len2), SPACE, len2, chemical_formula);
    ptr = "Formula Weight (g/mol)";
    len1 = strlen( ptr );
    fprintf( summary_file,"%*s%*c%7.2f\n", len1, ptr, LINE_LEN - (len1 + 7),
                          SPACE, formula_weight);
    ptr = "Crystal Dimensions (mm)";
    len1 = strlen( ptr );
    fprintf( summary_file, "%*s%*c%4.2f x %4.2f x %4.2f\n",len1, ptr,
                     LINE_LEN - (len1 + 18), SPACE, 
                     exper->crystal_size_max,
                     exper->crystal_size_med,
                     exper->crystal_size_min );
    ptr = "Crystal Family";
    len1 = strlen( ptr );
    len2 = strlen( latt->crystal_family );
    fprintf( summary_file,"%*s%*c%*s\n", len1, ptr,
             LINE_LEN - (len1 + len2), SPACE, len2, latt->crystal_family );
    ptr = "Space Group";
    len1 = strlen( ptr );
    len2 = strlen( latt->space_group );
    fprintf( summary_file,"%*s%*c%*s\n", len1, ptr, 
             LINE_LEN - (len1+len2), SPACE, len2, latt->space_group );
    ptr = "Temperature (K)";
    len1 = strlen( ptr );
    fprintf( summary_file,"%*s%*c%6.2f\n", len1, ptr, 
                          LINE_LEN -(len1 + 6), SPACE, exper->temperature );

    switch( crystal_family ) {
      int    prec_cell[6];   /* some local declarations */
      double sigma_cell[6];

      case TRICLINIC:
         prec_cell[0] = CalcSignificantFigures( unit->sigma_a, &sigma_cell[0] );
         prec_cell[1] = CalcSignificantFigures( unit->sigma_b, &sigma_cell[1] );
         prec_cell[2] = CalcSignificantFigures( unit->sigma_c, &sigma_cell[2] );
         prec_cell[3] = CalcSignificantFigures( unit->sigma_alpha, &sigma_cell[3] );
         prec_cell[4] = CalcSignificantFigures( unit->sigma_beta, &sigma_cell[4] );
         prec_cell[5] = CalcSignificantFigures( unit->sigma_gamma, &sigma_cell[5] );

         ptr = "a (Ang.)";
         len1 = strlen( ptr );
         fprintf( summary_file,"%*s%*c%7.*f(%2.0f)\n", len1, ptr, 
                  LINE_LEN -(len1+11), SPACE, prec_cell[0], unit->a, sigma_cell[0]);
         ptr = "b (Ang.)";
         len1 = strlen( ptr );
         fprintf( summary_file,"%*s%*c%7.*f(%2.0f)\n", len1, ptr, 
                  LINE_LEN -(len1+11), SPACE, prec_cell[1], unit->b, sigma_cell[1]);
         ptr = "c (Ang.)";
         len1 = strlen( ptr );
         fprintf( summary_file,"%*s%*c%7.*f(%2.0f)\n", len1, ptr, 
                  LINE_LEN -(len1+11), SPACE, prec_cell[2], unit->c, sigma_cell[2]);
         ptr = "alpha (deg)";
         len1 = strlen( ptr );
         fprintf( summary_file,"%*s%*c%7.*f(%2.0f)\n", len1, ptr, 
                  LINE_LEN -(len1+11), SPACE, prec_cell[3], unit->alpha, sigma_cell[3]);
         ptr = "beta (deg)";
         len1 = strlen( ptr );
         fprintf( summary_file,"%*s%*c%7.*f(%2.0f)\n", len1, ptr, 
                  LINE_LEN -(len1+11), SPACE, prec_cell[4], unit->beta, sigma_cell[4]);
         ptr = "gamma (deg)";
         len1 = strlen( ptr );
         fprintf( summary_file,"%*s%*c%7.*f(%2.0f)\n", len1, ptr, 
                  LINE_LEN -(len1+11), SPACE, prec_cell[5], unit->gamma, sigma_cell[5]);
         break;
      case MONOCLINIC:
         prec_cell[0] = CalcSignificantFigures( unit->sigma_a, &sigma_cell[0] );
         prec_cell[1] = CalcSignificantFigures( unit->sigma_b, &sigma_cell[1] );
         prec_cell[2] = CalcSignificantFigures( unit->sigma_c, &sigma_cell[2] );
         prec_cell[4] = CalcSignificantFigures( unit->sigma_beta, &sigma_cell[4] );
         ptr = "a (Ang.)";
         len1 = strlen( ptr );
         fprintf( summary_file,"%*s%*c%7.*f(%2.0f)\n", len1, ptr, 
                  LINE_LEN -(len1+11), SPACE, prec_cell[0], unit->a, sigma_cell[0]);
         ptr = "b (Ang.)";
         len1 = strlen( ptr );
         fprintf( summary_file,"%*s%*c%7.*f(%2.0f)\n", len1, ptr, 
                  LINE_LEN -(len1+11), SPACE, prec_cell[1], unit->b, sigma_cell[1]);
         ptr = "c (Ang.)";
         len1 = strlen( ptr );
         fprintf( summary_file,"%*s%*c%7.*f(%2.0f)\n", len1, ptr, 
                  LINE_LEN -(len1+11), SPACE, prec_cell[2], unit->c, sigma_cell[2]);
         ptr = "beta (deg)";
         len1 = strlen( ptr );
         fprintf( summary_file,"%*s%*c%7.*f(%2.0f)\n", len1, ptr, 
                  LINE_LEN -(len1+11), SPACE, prec_cell[4], unit->beta, sigma_cell[4]);
         break;
      case ORTHORHOMBIC:
         prec_cell[0] = CalcSignificantFigures( unit->sigma_a, &sigma_cell[0] );
         prec_cell[1] = CalcSignificantFigures( unit->sigma_b, &sigma_cell[1] );
         prec_cell[2] = CalcSignificantFigures( unit->sigma_c, &sigma_cell[2] );
         ptr = "a (Ang.)";
         len1 = strlen( ptr );
         fprintf( summary_file,"%*s%*c%7.*f(%2.0f)\n", len1, ptr, 
                  LINE_LEN -(len1+11), SPACE, prec_cell[0], unit->a, sigma_cell[0]);
         ptr = "b (Ang.)";
         len1 = strlen( ptr );
         fprintf( summary_file,"%*s%*c%7.*f(%2.0f)\n", len1, ptr, 
                  LINE_LEN -(len1+11), SPACE, prec_cell[1], unit->b, sigma_cell[1]);
         ptr = "c (Ang.)";
         len1 = strlen( ptr );
         fprintf( summary_file,"%*s%*c%7.*f(%2.0f)\n", len1, ptr, 
                  LINE_LEN -(len1+11), SPACE, prec_cell[2], unit->c, sigma_cell[2]);
         break;
      case TETRAGONAL:
         prec_cell[0] = CalcSignificantFigures( unit->sigma_a, &sigma_cell[0] );
         prec_cell[2] = CalcSignificantFigures( unit->sigma_c, &sigma_cell[2] );
         ptr = "a (Ang.)";
         len1 = strlen( ptr );
         fprintf( summary_file,"%*s%*c%7.*f(%2.0f)\n", len1, ptr, 
                  LINE_LEN -(len1+11), SPACE, prec_cell[0], unit->a, sigma_cell[0]);
         ptr = "c (Ang.)";
         len1 = strlen( ptr );
         fprintf( summary_file,"%*s%*c%7.*f(%2.0f)\n", len1, ptr, 
                  LINE_LEN -(len1+11), SPACE, prec_cell[2], unit->c, sigma_cell[2]);
         break;
      case RHOMBOHEDRAL:
         prec_cell[0] = CalcSignificantFigures( unit->sigma_a, &sigma_cell[0] );
         prec_cell[3] = CalcSignificantFigures( unit->sigma_alpha, &sigma_cell[3] );
         ptr = "a (Ang.)";
         len1 = strlen( ptr );
         fprintf( summary_file,"%*s%*c%7.*f(%2.0f)\n", len1, ptr, 
                  LINE_LEN -(len1+11), SPACE, prec_cell[0], unit->a, sigma_cell[0]);
         ptr = "alpha (deg)";
         len1 = strlen( ptr );
         fprintf( summary_file,"%*s%*c%7.*f(%2.0f)\n", len1, ptr, 
                  LINE_LEN -(len1+11), SPACE, prec_cell[3], unit->alpha, sigma_cell[3]);
         break;
      case TRIGONAL:  /* fall through */
      case HEXAGONAL:
         prec_cell[0] = CalcSignificantFigures( unit->sigma_a, &sigma_cell[0] );
         prec_cell[2] = CalcSignificantFigures( unit->sigma_c, &sigma_cell[2] );
         ptr = "a (Ang.)";
         len1 = strlen( ptr );
         fprintf( summary_file,"%*s%*c%7.*f(%2.0f)\n", len1, ptr, 
                  LINE_LEN -(len1+11), SPACE, prec_cell[0], unit->a, sigma_cell[0]);
         ptr = "c (Ang.)";
         len1 = strlen( ptr );
         fprintf( summary_file,"%*s%*c%7.*f(%2.0f)\n", len1, ptr, 
                  LINE_LEN -(len1+11), SPACE, prec_cell[2], unit->c, sigma_cell[2]);
         break;
      case CUBIC:
         prec_cell[0] = CalcSignificantFigures( unit->sigma_a, &sigma_cell[0] );
         ptr = "a (Ang.)";
         len1 = strlen( ptr );
         fprintf( summary_file,"%*s%*c%7.*f(%2.0f)\n", len1, ptr, 
                  LINE_LEN -(len1+11), SPACE, prec_cell[0], unit->a, sigma_cell[0]);
         break;
    }

/* now print out non crystal family specific data. */
   {               /*** LOCAL BLOCK ***/
     /* block specific data for getting the cell volume printed out. */
      double sigma_v;
      int    prec_v;
      prec_v = CalcSignificantFigures( unit->sigma_volume, &sigma_v );

     ptr = "V (A**3)";
     len1 = strlen( ptr );
     fprintf( summary_file,"%*s%*c%8.*f(%2.0f)\n",len1, ptr,
              LINE_LEN -(len1+12), SPACE, prec_v, unit->volume, sigma_v );
   }
    ptr = "Number of reflections to determine final unit cell";
    len1 = strlen( ptr );
    fprintf( summary_file,
             "%*s%*c%4hd\n",len1, ptr, LINE_LEN-(len1+4), SPACE,
                     exper->n_refl_unit_cell );
    ptr = "Min & Max 2(theta) values for cell determination (deg)";
    len1 = strlen( ptr );
    fprintf( summary_file,
            "%*s%*c%2.0f, %2.0f\n",len1, ptr, LINE_LEN-(len1+6), SPACE,
             exper->theta_min_unit_cell, exper->theta_max_unit_cell );
    ptr = "Z";
    len1 = strlen( ptr );
    fprintf( summary_file,"%*s%*c%4hd\n",len1, ptr, LINE_LEN-(len1+4),
                          SPACE, latt->n_mol_cell);
    ptr = "F(000)";
    len1 = strlen( ptr );
    fprintf( summary_file,"%*s%*c%7.2f\n",len1, ptr,
                          LINE_LEN-(len1+7), SPACE, exper->F000);
    ptr = "Density (g/cm**3)";
    len1 = strlen( ptr );
    fprintf( summary_file,"%*s%*c%6.3f\n", len1, ptr, LINE_LEN-(len1+6), 
                          SPACE, exper->crystal_density );
    { /* Local variable declaration */
    char *ptr_extra;
    ptr = "Wavelength (";
    ptr_extra = "lpha)";
    len1 = strlen( ptr );
    len2 = strlen( ptr_extra );
    fprintf( summary_file,"%*s%2s%*s%*c%7.5f\n", len1, ptr, unit->radiation, len2,
                           ptr_extra, LINE_LEN-(len1+len2+11), SPACE,
                           unit->wavelength );

    }
    ptr = "mu (cm**-1)";
    len1 = strlen( ptr );
    fprintf( summary_file,"%*s%*c%6.2f\n", len1, ptr, LINE_LEN-(len1+6),SPACE, exper->mu);
    { /* local variable declaration */
    size_t len3;
    ptr = "Diffractometer Type";
    len1 = strlen( ptr );
    len2 = strlen( exper->vendor_id ) + 1;
    len3 = strlen( exper->model_id ) + 1;
    fprintf( summary_file,"%*s%*c%*s%*s\n",len1, ptr, LINE_LEN-(len1+len2+len3), SPACE, len2,
                     exper->vendor_id, len3, exper->model_id);
    }
    ptr = "Data Collection";
    len1 = strlen( ptr );
    len2 = strlen( exper->scan_mode );
    fprintf( summary_file, "%*s%*c%*s scan mode\n", len1, ptr, LINE_LEN-(len1+len2+10),
                           SPACE, len2, exper->scan_mode);
    ptr = "Max 2(theta) for data collection (deg)";
    len1 = strlen( ptr );
    fprintf( summary_file,"%*s%*c%4.1f\n",len1, ptr, LINE_LEN-(len1+4), SPACE,
                          2.0 * exper->theta_max);
    ptr = "Number of Reflections measured";
    len1 = strlen( ptr );
    fprintf( summary_file,"%*s%*c%4d\n", len1, ptr, LINE_LEN-(len1+4), SPACE,
                          exper->reflns_measured);

    ptr = "Unique Reflections Measured";
    len1 = strlen( ptr );
    fprintf( summary_file,"%*s%*c%4d\n",len1, ptr, LINE_LEN-(len1+4), SPACE,
                          exper->reflns_unique );
    ptr = "Rmerge";
    len1 = strlen( ptr );
    fprintf( summary_file,"%*s%*c%5.3f\n",len1, ptr, LINE_LEN-(len1+5), SPACE,
                          exper->r_merge );

    if( RelativeDifference(exper->transmission_min, exper->transmission_max) ) {
       ptr = "min & max transmission";
       len1 = strlen( ptr );
       fprintf( summary_file,"%*s%*c%5.3f, %5.3f\n", len1, ptr, LINE_LEN-(len1+12), SPACE, 
                             exper->transmission_min,
                             exper->transmission_max );
    }
    else
       fprintf( summary_file,"Data not corrected for absorption\n");

    ptr = "Number of Reflections included in refinement";
    len1 = strlen( ptr );
    fprintf( summary_file,"%*s%*c%4d\n",len1, ptr, LINE_LEN-(len1+4), SPACE,
                          refine->refine_reflns_incl );

    ptr = "Cut off threshold";
    len1 = strlen( ptr );
    fprintf( summary_file,"%*s%*cInet > %3.1fsigma(I)\n",len1, ptr, 
                          LINE_LEN-(len1+18), SPACE, exper->sigma_cutoff);

    fprintf( summary_file, "Structure Refined Using %s matrix Least Squares Method based on %s\n",
                           refine->matrix_type, refine->refine_type );
    if( 0 != strcmp(refine->weight_scheme, "unit")) {
       fprintf( summary_file,"Weighting Scheme based on %s\n", refine->weight_scheme);
       fprintf( summary_file,"Where weight = 1 /(sigma(Fo)**2 + %7.4f * Fo**2)\n", refine->weight_coeff);
    }
    else
       fprintf( summary_file,"Structure refined with Unit Weights\n");
    
    ptr = "Number of Parameters in Least-Squares";
    len1 = strlen( ptr );
    fprintf( summary_file,"%*s%*c%4d\n",len1, ptr, LINE_LEN-(len1+4), SPACE, refine->n_parameters);
    ptr = "R";
    len1 = strlen( ptr );
    fprintf( summary_file,"%*s%*c%5.3f\n",len1, ptr, LINE_LEN-(len1+5), SPACE, refine->r_conventional);
    ptr = "Rw";
    len1 = strlen( ptr );
    fprintf( summary_file,"%*s%*c%5.3f\n",len1, ptr, LINE_LEN-(len1+5), SPACE, refine->r_weighted);
    ptr = "GOF";
    len1 = strlen( ptr );
    fprintf( summary_file,"%*s%*c%5.3f\n",len1, ptr, LINE_LEN-(len1+5), SPACE, refine->gof);
    ptr = "Maximum shift / error";
    len1 = strlen( ptr );
    fprintf( summary_file,"%*s%*c%5.3f\n",len1, ptr, LINE_LEN-(len1+5), SPACE, refine->shift_err);
    ptr = "Min & Max Peak Heights on Final DF Map (e-/A**3)";
    len1 = strlen( ptr );
    fprintf( summary_file,"%*s%*c%5.2f, %4.2f\n",len1, ptr, LINE_LEN-(len1+11), SPACE,
                          refine->final_df_min, refine->final_df_max );
    fprintf( summary_file, "\n");
    fprintf( summary_file,"Where: \n");
    fprintf( summary_file,"Rf = Sum( Fo - Fc ) / Sum( Fo )\n");
    fprintf( summary_file,"Rw = Sqrt[ Sum( w( Fo - Fc )**2 ) / Sum( Fo**2 ) ]\n");
    fprintf( summary_file,"GOF = Sqrt[ Sum( w( Fo - Fc )**2 ) / (No. of reflns. - No. of params. ) ]\n");


    if( CLI_MODE != run_mode ) 
        fprintf( summary_file, "</pre></body></html>\n" );

    fclose( summary_file );
    free( chemical_formula );
    return;
}

/* this function outputs the xyz's and ADP's in separate files */
void ProduceAtomParamsFile( char *struct_name, AtomCoords *atoms,
                            UnitCell *cell,  int n_atoms )
{
    FILE *xyz_file,
         *adp_file;

    char *xyz_file_name,
         *adp_file_name,
         *xyz_suffix,
         *adp_suffix,
         *xyz_title,
         *adp_title;
 
     char error_msg[80];
     int  i_atom,
          i;

    if( CLI_MODE == run_mode ) {
      xyz_suffix = XYZ_SUFFIX_TXT;
      adp_suffix = ADP_SUFFIX_TXT;
    }
    else {
      xyz_suffix = XYZ_SUFFIX_HTML;
      adp_suffix = ADP_SUFFIX_HTML;
    }

/* build up the output file names and open them. */

    xyz_file_name = malloc( strlen(xyz_suffix) + strlen(struct_name) + 1 );
    if( !xyz_file_name ) {
      if( CLI_MODE == run_mode )
          fprintf( stderr, "Memory allocation: %s:%d\n", 
                         __FILE__, __LINE__ );
      else
          ReportErrorHTML( "Memory Allocation" );

      exit( EXIT_FAILURE );
    }
    sprintf( xyz_file_name, xyz_suffix, struct_name );

    adp_file_name = malloc( strlen(adp_suffix) + strlen(struct_name) + 1 );
    if( !adp_file_name ) {
      if( CLI_MODE == run_mode )
          fprintf( stderr, "Memory allocation: %s:%d\n", 
                         __FILE__, __LINE__ );
      else
          ReportErrorHTML( "Memory Allocation" );

      exit( EXIT_FAILURE );
    }
    sprintf( adp_file_name, adp_suffix, struct_name );

    xyz_file = fopen( xyz_file_name, "w" );
    if( !xyz_file ) {
      if( CLI_MODE == run_mode ) {
           fprintf( stderr, "fopen(): %s:%d\n", 
                    __FILE__, __LINE__ );
           perror( xyz_file_name );
      }
      else {
           sprintf( error_msg, "Can't open %s: %s\n", xyz_file_name, 
                    strerror(errno) );
           ReportErrorHTML( error_msg );
      }

      exit( EXIT_FAILURE );
    }

    adp_file = fopen( adp_file_name, "w" );
    if( !adp_file ) {
      if( CLI_MODE == run_mode ) {
           fprintf( stderr, "fopen(): %s:%d\n", 
                    __FILE__, __LINE__ );
           perror( adp_file_name );
      }
      else {
           sprintf( error_msg, "Can't open %s: %s\n", adp_file_name, 
                    strerror(errno) );
           ReportErrorHTML( error_msg );
      }

      exit( EXIT_FAILURE );
    }

/* Build up the title strings */

    xyz_title = malloc( strlen(XYZ_TITLE) + strlen(struct_name) + 1 );
    if( !xyz_title ) {
      if( CLI_MODE == run_mode )
          fprintf( stderr, "Memory allocation: %s:%d\n", 
                         __FILE__, __LINE__ );
      else
          ReportErrorHTML( "Memory Allocation" );

      exit( EXIT_FAILURE );
    }
    sprintf( xyz_title, XYZ_TITLE, struct_name );

    adp_title = malloc( strlen(ADP_TITLE) + strlen(struct_name) + 1 );
    if( !adp_title ) {
      if( CLI_MODE == run_mode )
          fprintf( stderr, "Memory allocation: %s:%d\n", 
                         __FILE__, __LINE__ );
      else
          ReportErrorHTML( "Memory Allocation" );

      exit( EXIT_FAILURE );
    }
    sprintf( adp_title, ADP_TITLE, struct_name );


/* Write the standard html headers to the files along with the titles */
    if( CLI_MODE == run_mode ) {
        fprintf( xyz_file, "%s\n\n", xyz_title );
        fprintf( adp_file, "%s\n\n", adp_title );
    }
    else {
       fprintf( xyz_file, STD_HTML_HEADER, xyz_title, xyz_title );
       fprintf( adp_file, STD_HTML_HEADER, adp_title, adp_title );

       fprintf( xyz_file, "<pre>\n" );
       fprintf( adp_file, "<pre>\n" );
    }

#ifdef USE_B_TERMS
    fprintf(xyz_file,"Atom          x               y               z         B(iso)\n");
#else
    fprintf(xyz_file,"Atom          x               y               z         U(iso)\n");
#endif

fprintf(adp_file,"Atom         u11             u22             u33             u12             u13             u23\n");

/* Loop through the atom records and write the information to
 * the proper files.
 */
    for( i_atom = 0; i_atom < n_atoms; i_atom++ ) {
      double   sigma_xyz[3],
               sigma_adp[6],
             new_sig_adp[6],
                     adp[6],
              u_equiv = 0.0;     

      /* prec_xyz[] is the number of sig figs.  If a coordinate has
       * been fixed the default precision is 5 places.  Similarly,
       * for ADP's the default precision is 4 places.
       */
      int    prec_xyz[3] = { 5, 5, 5 },     
             prec_adp[6] = { 4, 4, 4, 4, 4, 4 };
     

       /* assign some local variables to clean up clutter. */
       adp[0] = atoms[i_atom].u11;
       adp[1] = atoms[i_atom].u22;
       adp[2] = atoms[i_atom].u33;
       adp[3] = atoms[i_atom].u12;
       adp[4] = atoms[i_atom].u13;
       adp[5] = atoms[i_atom].u23;

       sigma_adp[0] = atoms[i_atom].sigma_u11;
       sigma_adp[1] = atoms[i_atom].sigma_u11;
       sigma_adp[2] = atoms[i_atom].sigma_u33;
       sigma_adp[3] = atoms[i_atom].sigma_u12;
       sigma_adp[4] = atoms[i_atom].sigma_u13;
       sigma_adp[5] = atoms[i_atom].sigma_u23;
      
       if( IsAnisotropic(atoms[i_atom].refine_flags) )
           u_equiv = CalcEquivADP( atoms[i_atom].u11, atoms[i_atom].u22,
                                   atoms[i_atom].u33, atoms[i_atom].u12,
                                   atoms[i_atom].u13, atoms[i_atom].u23 );
       else
          u_equiv = atoms[i_atom].u11;

     switch( PositionalRefinementFlags(atoms[i_atom].refine_flags) )
     {
       case 0:  /* no coordinates are fixed. */
          prec_xyz[0] = CalcSignificantFigures( atoms[i_atom].sigma_x,
                                                &sigma_xyz[0] );
          prec_xyz[1] = CalcSignificantFigures( atoms[i_atom].sigma_y,
                                                &sigma_xyz[1] );
          prec_xyz[2] = CalcSignificantFigures( atoms[i_atom].sigma_z,
                                                &sigma_xyz[2] );
          fprintf( xyz_file,"%-6s\t%9.*f(%2.0f)\t%9.*f(%2.0f)\t%9.*f(%2.0f)\t%5.2f\n",
                           atoms[i_atom].atom_name,
                           prec_xyz[0], atoms[i_atom].x,sigma_xyz[0],
                           prec_xyz[1], atoms[i_atom].y,sigma_xyz[1],
                           prec_xyz[2], atoms[i_atom].z,sigma_xyz[2],
                           u_equiv );

          break;
       case 1:  /* x is fixed */
          prec_xyz[1] = CalcSignificantFigures( atoms[i_atom].sigma_y,
                                                &sigma_xyz[1] );
          prec_xyz[2] = CalcSignificantFigures( atoms[i_atom].sigma_z,
                                                &sigma_xyz[2] );
          fprintf(xyz_file,"%-6s\t%9.*f\t%9.*f(%2.0f)\t%9.*f(%2.0f)\t%5.2f\n",
                           atoms[i_atom].atom_name,
                           prec_xyz[0], atoms[i_atom].x,
                           prec_xyz[1], atoms[i_atom].y,sigma_xyz[1],
                           prec_xyz[2], atoms[i_atom].z,sigma_xyz[2],
                           u_equiv );
          break;
       case 2: /* y is fixed */
          prec_xyz[0] = CalcSignificantFigures( atoms[i_atom].sigma_x,
                                                &sigma_xyz[0] );
          prec_xyz[2] = CalcSignificantFigures( atoms[i_atom].sigma_z,
                                                &sigma_xyz[2] );
          fprintf(xyz_file,"%-6s\t%9.*f(%2.0f)\t%9.*f\t%9.*f(%2.0f)\t%5.2f\n",
                           atoms[i_atom].atom_name,
                           prec_xyz[0], atoms[i_atom].x,sigma_xyz[0],
                           prec_xyz[1], atoms[i_atom].y,
                           prec_xyz[2], atoms[i_atom].z,sigma_xyz[2],
                           u_equiv );
          break;
       case 3: /* x and  y are fixed */
          prec_xyz[2] = CalcSignificantFigures( atoms[i_atom].sigma_z,
                                                &sigma_xyz[2] );
          fprintf(xyz_file,"%-6s\t%9.*f\t%9.*f\t%9.*f(%2.0f)\t%5.2f\n",
                           atoms[i_atom].atom_name,
                           prec_xyz[0], atoms[i_atom].x,
                           prec_xyz[1], atoms[i_atom].y,
                           prec_xyz[2], atoms[i_atom].z,sigma_xyz[2],
                           u_equiv );
          break;
       case 4: /* z is fixed */
          prec_xyz[0] = CalcSignificantFigures( atoms[i_atom].sigma_x,
                                                &sigma_xyz[0] );
          prec_xyz[1] = CalcSignificantFigures( atoms[i_atom].sigma_y,
                                                &sigma_xyz[1] );
          fprintf(xyz_file,"%-6s\t%9.*f(%2.0f)\t%9.*f(%2.0f)\t%9.*f\t%5.2f\n",
                           atoms[i_atom].atom_name,
                           prec_xyz[0], atoms[i_atom].x,sigma_xyz[0],
                           prec_xyz[1], atoms[i_atom].y,sigma_xyz[1],
                           prec_xyz[2], atoms[i_atom].z,
                           u_equiv );
          break;
       case 5: /* x and z are fixed */
          prec_xyz[1] = CalcSignificantFigures( atoms[i_atom].sigma_y,
                                                &sigma_xyz[1] );
          fprintf(xyz_file,"%-6s\t%9.*f\t%9.*f(%2.0f)\t%9.*f\t%5.2f\n",
                           atoms[i_atom].atom_name,
                           prec_xyz[0], atoms[i_atom].x,
                           prec_xyz[1], atoms[i_atom].y,sigma_xyz[1],
                           prec_xyz[2], atoms[i_atom].z,
                           u_equiv );
          break;
       case 6: /* y and z are fixed */
          prec_xyz[0] = CalcSignificantFigures( atoms[i_atom].sigma_x,
                                                &sigma_xyz[0] );
          fprintf(xyz_file,"%-6s\t%9.*f(%2.0f)\t%9.*f\t%9.*f\t%5.2f\n",
                           atoms[i_atom].atom_name,
                           prec_xyz[0], atoms[i_atom].x,sigma_xyz[0],
                           prec_xyz[1], atoms[i_atom].y,
                           prec_xyz[2], atoms[i_atom].z,
                           u_equiv );
          break;
       case 7: /* x, y, and  z are fixed */
          fprintf(xyz_file,"%-6s\t%9.*f\t%9.*f\t%9.*f\t%5.2f\n",
                           atoms[i_atom].atom_name,
                           prec_xyz[0], atoms[i_atom].x,
                           prec_xyz[1], atoms[i_atom].y,
                           prec_xyz[2], atoms[i_atom].z,
                           u_equiv );
          break;
     }   /* end of positional xyz switch() */

     /* now process the ADP's  the precision for fixed Uij terms was initialized
      * at the top of the for( i_atom ) loop
      */
     if( IsAnisotropic( atoms[i_atom].refine_flags ) ) {
         for( i = 0; i < 6; i++ ) 
           if( !RelativeDifference(adp[i], 0.0) <= TOLERANCE ) { 
             prec_adp[i] = CalcSignificantFigures( sigma_adp[i], &new_sig_adp[i] );
           }
           fprintf( adp_file, "%-6s\t%8.*f(%2.0f)\t%8.*f(%2.0f)\t%8.*f(%2.0f)"
                        "\t%8.*f(%2.0f)\t%8.*f(%2.0f)\t%8.*f(%2.0f)\n",
                       atoms[i_atom].atom_name,
                       prec_adp[0], adp[0], new_sig_adp[0],
                       prec_adp[1], adp[1], new_sig_adp[1],
                       prec_adp[2], adp[2], new_sig_adp[2],
                       prec_adp[3], adp[3], new_sig_adp[3],
                       prec_adp[4], adp[4], new_sig_adp[4],
                       prec_adp[5], adp[5], new_sig_adp[5] );

           
     }
     else if( IsAnisotropic(atoms[i_atom].refine_flags) &&
              (atoms[i_atom].refine_flags & (1<<5)) ) { /* for isotropic atoms
                                                         if both expression return 0
                                                         then isotropic atom was refined */
 
        prec_adp[0] = CalcSignificantFigures( sigma_adp[0], &new_sig_adp[0] );
        fprintf( adp_file, "%-6s\t%8.*f(%2.0f)\n",
                       atoms[i_atom].atom_name,
                       prec_adp[0], adp[0],
                       new_sig_adp[0] );

     }
     else /* unrefined isotropic atoms */
        fprintf( adp_file, "%-6s\t%8.*f\n", atoms[i_atom].atom_name, prec_adp[0], adp[0] );

       
    }  /* end of for( i_atom ) loop */

#ifdef USE_B_TERMS
    fprintf( xyz_file, "\n\nB(iso) is 8*Pi*Pi times the average of the diagonal elements of the ADP's\n");
#else
    fprintf( xyz_file, "\n\nU(iso) is the average of the diagonal elements of the ADP's\n");
#endif    
    fprintf( adp_file, "\n\nAnisotropic Displacement Parameters (ADP's) are of the form\n" );
    fprintf( adp_file, "ADP = 2*Pi*Pi*(h*h*u11*astar*astar + ... + 2*h*k*u12*astar*bstar + ... )\n" );    

    if( CGI_MODE == run_mode ) {
        fprintf( xyz_file, "</pre>\n" );
        fprintf( adp_file, "</pre>\n" );

        fprintf( xyz_file, STD_HTML_FOOTER );
        fprintf( adp_file, STD_HTML_FOOTER );
    }

    fclose( xyz_file );
    fclose( adp_file );

}
/* this function writes a file of the specified geometric
 * quantities.  Usage:
 *   MolecularGeometryTable( "x97123", bonds, 35, BOND_LENGTH );
 */
void MolecularGeometryTable( char *struct_name, void *geom,
                              int n_items, enum TableType tab_type )
{
    FILE *out_file;

    char *out_file_name,
         *out_title, 
         *suffix_ptr,
         *title_ptr;
 
     char error_msg[80];
     int  i_item,
          k_rows,
          is_odd;  /* 1 for ODD, 0 for EVEN */

     BondLength *bonds = NULL;
     BondAngle  *angles = NULL;

/* cast the geom pointer and set up the title and suffix. */

     switch( tab_type ) {
       case BOND_LENGTH:
          bonds = (BondLength *)geom;
          if( CLI_MODE == run_mode )
              suffix_ptr = BOND_SUFFIX_TXT;
          else
              suffix_ptr = BOND_SUFFIX_HTML;
          title_ptr  = BOND_TITLE;
          break;
       case BOND_ANGLE:
          angles = (BondAngle *)geom;
          if( CLI_MODE == run_mode )
              suffix_ptr = ANGLE_SUFFIX_TXT;
          else
              suffix_ptr = ANGLE_SUFFIX_HTML;
          title_ptr  = ANGLE_TITLE;
          break;
       default:
          if( CLI_MODE == run_mode )
              fputs( "MolGeom: Bad Table Type Specifier\n", stderr );
          else
              ReportErrorHTML( "MolGeom: Bad Table Type Specifier\n" );
          return;
          break;
     }

/* build up the output file names and open them. */

    out_file_name = malloc( strlen(suffix_ptr) + strlen(struct_name) + 1 );
    if( !out_file_name ) {
        if( CLI_MODE == run_mode )
            fprintf( stderr, "Memory allocation: %s:%d\n", 
                             __FILE__, __LINE__ );
        else
            ReportErrorHTML( "Memory Allocation" );

        exit( EXIT_FAILURE );
    }
    sprintf( out_file_name, suffix_ptr, struct_name );


    out_file = fopen( out_file_name, "w" );
    if( !out_file ) {
       if( CLI_MODE == run_mode ) {
          fprintf( stderr, "fopen(): %s:%d\n", __FILE__, __LINE__ );
          perror( out_file_name );
       }
       else {
          sprintf( error_msg, "Can't open %s: %s\n", out_file_name, 
                   strerror(errno) );
          ReportErrorHTML( error_msg );
       } 
       exit( EXIT_FAILURE );
    }


/* Build up the title strings */

    out_title = malloc( strlen(title_ptr) + strlen(struct_name) + 1 );
    if( !out_title ) {
        if( CLI_MODE == run_mode )
            fprintf( stderr, "Memory allocation: %s:%d\n", 
                             __FILE__, __LINE__ );
        else
            ReportErrorHTML( "Memory Allocation" );

        exit( EXIT_FAILURE );
    }
    sprintf( out_title, title_ptr, struct_name );


/* Write the standard html headers to the files along with the titles */
    if( CLI_MODE == run_mode )
       fprintf( out_file, "%s\n\n", out_title );
    else {
       fprintf( out_file, STD_HTML_HEADER, out_title, out_title );
       fprintf( out_file, "<pre>\n" );
    }

/* Loop through the atom records and write the information to
 * the proper files. Make the "table" 2 columns wide. (see #define N_COLS ).
 * If more columns are desired, this will have to be rewritten.
 */
    if( n_items % N_COLS ) {
        is_odd = 1;
        k_rows = (n_items + 1 ) / N_COLS;  /* odd number of items */
    }
    else {
        is_odd = 0;
        k_rows = n_items / N_COLS;         /* even number of items */
    }

    for( i_item = 0; i_item < k_rows; i_item++ ) {
       double sigma[N_COLS];
          int prec[N_COLS];
       switch( tab_type ) {
         case BOND_LENGTH: { /* local variables */
            char label_col_1[16];
            char label_col_2[16];
             
            if( (0 == is_odd) || (i_item < k_rows - 1) ) {
               sprintf( label_col_1, "%s-%s", bonds[i_item].atom_1, bonds[i_item].atom_2 );
               sprintf( label_col_2, "%s-%s", bonds[i_item + k_rows].atom_1,
                                      bonds[i_item + k_rows].atom_2 );
   
               prec[0] = CalcSignificantFigures( bonds[i_item].esd, &sigma[0] );
               prec[1] = CalcSignificantFigures( bonds[i_item + k_rows].esd, &sigma[1] );
               fprintf( out_file, "%-16s%8.*f(%2.0f)\t%-16s%8.*f(%2.0f)\n",
                        label_col_1, prec[0],
                        bonds[i_item].dist, sigma[0],
                        label_col_2, prec[1],
                        bonds[i_item + k_rows].dist, sigma[1] );
            }
            else {
               sprintf( label_col_1, "%s-%s", bonds[i_item].atom_1, bonds[i_item].atom_2 );
               prec[0] = CalcSignificantFigures( bonds[i_item].esd, &sigma[0] );
               fprintf( out_file, "%-16s%8.*f(%2.0f)\n",
                        label_col_1, prec[0],
                        bonds[i_item].dist, sigma[0] );
            }
         }
            break;
         case BOND_ANGLE: {
            char label_col_1[20];
            char label_col_2[20];

            if( (0 == is_odd) || (i_item < k_rows - 1) ) {
               sprintf( label_col_1, "%s-%s-%s", angles[i_item].atom_1,
                                     angles[i_item].atom_2, angles[i_item].atom_3 );
               sprintf( label_col_2, "%s-%s-%s", angles[i_item + k_rows].atom_1,
                                      angles[i_item + k_rows].atom_2,
                                      angles[i_item + k_rows].atom_3 );
   
               prec[0] = CalcSignificantFigures( angles[i_item].esd, &sigma[0] );
               prec[1] = CalcSignificantFigures( angles[i_item + k_rows].esd, &sigma[1] );
               fprintf( out_file, "%-20s%8.*f(%2.0f)\t%-20s%8.*f(%2.0f)\n",
                        label_col_1,
                        prec[0],
                        angles[i_item].angle, sigma[0],
                        label_col_2,
                        prec[1],
                        angles[i_item + k_rows].angle, sigma[1] );
            }
            else {
               sprintf( label_col_1, "%s-%s-%s", angles[i_item].atom_1,
                                     angles[i_item].atom_2, angles[i_item].atom_3 );
               prec[0] = CalcSignificantFigures( angles[i_item].esd, &sigma[0] );
               fprintf( out_file, "%-20s%8.*f(%2.0f)\n",
                        label_col_1,
                        prec[0],
                        angles[i_item].angle, sigma[0] );
            }
          }
            break;
          default: /* should never get here */
             break;
       }
    }
    
    if( CGI_MODE == run_mode ) {
        fprintf( out_file, "</pre>\n" );
        fprintf( out_file, STD_HTML_FOOTER );
    }

    fclose( out_file );
    return;
}
