/* this file contains the functions for converting the Tuple struct's
 * into crystallographic data struct's.
 * Written by:
 *	Paul D. Boyle
 *	Department of Chemistry
 *	Box 8204
 *	North Carolina State University
 *	Raleigh, NC 27695-8204
 *
 *	March 1998
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#define _XOPEN_SOURCE

#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#ifndef HAVE_SINFO_H
#define HAVE_SINFO_H 1
#include "sinfo.h"
#endif
#include "qthread.h"
#include "periodic_table.h"
#include "process_tuples.h"

Lattice *ProcessLatticeTuple( Tuple *tup, int n_tup, int n_field )
{
    int h = 0,
        i = 1,
        j = 0;

    Lattice *latt_ptr;

/* these lattice item names are as they have been defined in the database.
 * If the datanbase item names change, then these names must change also.
 */
    static char *lattice_item_names[] = {  "unit_cell_seq",
                                           "n_mol_cell",
                                           "crystal_family",
                                           "lattice_type",
                                           "space_group", 
                                           "" };
    latt_ptr = malloc( n_tup * sizeof(Lattice) );
    if( !latt_ptr ) 
        return NULL;
    
    for( h = 0; h < n_tup; h++ ) {
       i = 0;
       while( 0 != strcmp(lattice_item_names[i], "") ) {
          for( j = 0; j < n_field; j++ ) {
             if( 0 == strcmp(lattice_item_names[i], tup[h].item_name[j]) ) {
                 switch( i ) {
                   case 1:  /* n_mol_cell */
                      latt_ptr[h].n_mol_cell = (short)strtol(tup[h].value[j],
                                                     (char **)NULL, 10);
                      break;
                   case 2:
                      memset( latt_ptr[h].crystal_family, '\0', CRYST_FAM_LEN );
                      strcpy( latt_ptr[h].crystal_family, tup[h].value[j] );
                      break;
                   case 3:
                      latt_ptr[h].lattice_type = tup[h].value[j][0];
                      break;
                   case 4:
                      memset( latt_ptr[h].space_group, '\0', SPACE_GROUP_LEN );
                      strcpy( latt_ptr[h].space_group, tup[h].value[j] );
                      break;
                 } /* end switch() */
             }     /* end if(strcmp()) */
          }        /* end of for(n_field) loop */
          i++;
       }           /* end while() */
    }
    return latt_ptr;
                             
}

UnitCell *ProcessUnitCellTuple( Tuple *tup, int n_tup, int n_field )
{
    int h = 0,
        i = 1,
        j = 0;

    UnitCell *unit_cell_ptr;
    static char *unit_cell_item_names[] = { "unit_cell_seq",
                                          "id",
                                          "sample_date",
                                          "principal",
                                          "sample_orig", 
                                          "radiation", 
                                          "wavelength",
                                          "a", "b", "c", "alpha",
                                          "beta", "gamma", "volume",
                                          "reduced_vol", "sigma_a",
                                          "sigma_b", "sigma_c", 
                                          "sigma_alpha", "sigma_beta", 
                                          "sigma_gamma", "" };  

   unit_cell_ptr = malloc( n_tup * sizeof(UnitCell) );
   if( !unit_cell_ptr ) 
       return NULL;
    for( h = 0; h < n_tup; h++ ) {
       i = 0;
       while( 0 != strcmp(unit_cell_item_names[i], "") ) {
         for( j = 0; j < n_field; j++ ) {
           if( 0 == strcmp(unit_cell_item_names[i], tup[h].item_name[j]) ) {
               switch( i ) {
                 case 1:
                    memset( unit_cell_ptr[h].id, '\0', ID_LEN );
                    strcpy( unit_cell_ptr[h].id, tup[h].value[j] );
                    break;
                 case 2:
                    memset( unit_cell_ptr[h].sample_date, '\0', DATE_LEN );
                    strcpy( unit_cell_ptr[h].sample_date, tup[h].value[j] );
                    break;
                 case 3:
                    memset( unit_cell_ptr[h].principal, '\0', NAME_LEN );
                    strcpy( unit_cell_ptr[h].principal, tup[h].value[j] );
                    break;
                 case 4:
                    memset( unit_cell_ptr[h].sample_orig, '\0', NAME_LEN );
                    strcpy( unit_cell_ptr[h].sample_orig, tup[h].value[j] );
                    break;
                 case 5:
                    memset( unit_cell_ptr[h].radiation,'\0', RAD_LEN );
                    strcpy( unit_cell_ptr[h].radiation, tup[h].value[j] );
                    break;
                 case 6:
                    unit_cell_ptr[h].wavelength = strtod( tup[h].value[j], (char **)NULL );
                    break;
                 case 7:
                    unit_cell_ptr[h].a = strtod( tup[h].value[j], (char **)NULL );
                    break;
                 case 8:
                    unit_cell_ptr[h].b = strtod( tup[h].value[j], (char **)NULL );
                    break;
                 case 9:
                    unit_cell_ptr[h].c = strtod( tup[h].value[j], (char **)NULL );
                    break;
                 case 10:
                    unit_cell_ptr[h].alpha = strtod( tup[h].value[j], (char **)NULL );
                    break;
                 case 11:
                    unit_cell_ptr[h].beta = strtod( tup[h].value[j], (char **)NULL );
                    break;
                 case 12:
                    unit_cell_ptr[h].gamma = strtod( tup[h].value[j], (char **)NULL );
                    break;
                 case 13:
                    unit_cell_ptr[h].volume = strtod( tup[h].value[j], (char **)NULL );
                    break;
                 case 14:
                    unit_cell_ptr[h].reduced_vol = strtod( tup[h].value[j], (char **)NULL );
                    break;
                 case 15:
                    unit_cell_ptr[h].sigma_a = strtod( tup[h].value[j], (char **)NULL );
                    break;
                 case 16:
                    unit_cell_ptr[h].sigma_b = strtod( tup[h].value[j], (char **)NULL );
                    break;
                 case 17:
                    unit_cell_ptr[h].sigma_c = strtod( tup[h].value[j], (char **)NULL );
                    break;
                 case 18:
                    unit_cell_ptr[h].sigma_alpha = strtod( tup[h].value[j], (char **)NULL );
                    break;
                 case 19:
                    unit_cell_ptr[h].sigma_beta = strtod( tup[h].value[j], (char **)NULL );
                    break;
                 case 20:
                    unit_cell_ptr[h].sigma_gamma = strtod( tup[h].value[j], (char **)NULL );
                    break;
               } /* end of switch() */   
           } /* end if() */
         }  /* end for() */
         i++;
       }     /* end while() */
    }
    return unit_cell_ptr;
}

Formula *ProcessFormulaTuple( Tuple *tup, int n_tup, int n_field )
{
    int h = 0,
        i = 1,
        j = 0;
    short atomic_number;
 
    Formula *form_ptr;
    static char *formula_items[] = { "unit_cell_seq",
                                     "atomic_number",
                                     "count",
                                     "" };
    form_ptr = malloc( n_tup * sizeof(Formula) );
    if( !form_ptr ) 
        return NULL;

    for( h = 0; h < n_tup; h++ ) {
       i = 0;
       while( 0 != strcmp(formula_items[i], "") ) {
         for( j = 0; j < n_field; j++ ) {
           if( 0 == strcmp(formula_items[i], tup[h].item_name[j]) ) {
               switch( i ) {
                 case 1: /* periodic_table[] is defined in periodic_table.h file */
                    atomic_number = (short)strtol(tup[h].value[j],
                                                 (char **)NULL, 10);
                    form_ptr[h].element = periodic_table[atomic_number];
                    break;
                 case 2:
                    form_ptr[h].count = strtod( tup[h].value[j], (char **)NULL );
                    break;
               }
           }
         }
         i++;
       }
    }
    return form_ptr;
    
}

Experimental *ProcessExperimentalTuple( Tuple *tup, int n_tup, int n_field )
{
    int h = 0,
        i = 1,
        j = 0,
        k = 0;

     int n_instr,
         n_scans,
         n_vend;
       
     int   model_number,
           scan_number,
           vendor_number;

    Experimental *exper_ptr;

    struct instrument instr[] = { 
                      {HOME_MADE, "Contructed Locally"},
                      {CAD4, "CAD4"},
                      {CAD4_MACH, "CAD4-MACH"},
                      {CAD4_TURBO, "CAD4-Turbo"}, 
                      {MACH3, "Mach3"},
                      {KCCD, "KCCD"},
                      {P4, "P4"},
                      {P3, "P3"},
                      {R3, "R3"},
                      {SMART_CCD, "SMART CCD"},
                      {P21, "P21"},
                      {KM_4, "KM-4"},
                      {AFC5, "AFC5"},
                      {AFC6, "AFC6"},
                      {AFC7, "AFC7"} };

    struct vendor vend[] = {
                  {LOCAL, ""},
                  {NONIUS, "Nonius"},
                  {ENRAF_NONIUS, "Enraf-Nonius"},
                  {SIEMENS, "Siemens"},
                  {BRUKER, "Bruker"},
                  {RIGAKU, "Rigaku"},
                  {KUMA, "Kuma" } };

    struct scan_type scans[] = {
           {DUMMY, "Huh?"},
           {NONE, "None"},
           {THETA_2THETA, "theta/2theta"},
           {OMEGA, "omega"},
           {SPECIAL, "special"} };


    static char *exper_items[] = { "unit_cell_seq", "scan_mode", "vendor_id", "model_id",
                                   "temperature", "n_refl_unit_cell", "theta_min_unit_cell",
                                   "theta_max_unit_cell", "crystal_size_max",
                                   "crystal_size_med", "crystal_size_min", "crystal_density",
                                   "f000", "mu", "transmission_min", "transmission_max",
                                   "reflns_measured", "reflns_unique", "reflns_observed",
                                   "sigma_cutoff", "theta_min", "theta_max", "r_merge",
                                   "" };
    n_instr = sizeof instr / sizeof( struct instrument );
    n_vend = sizeof vend / sizeof( struct vendor );
    n_scans = sizeof scans / sizeof( struct scan_type );
 

    exper_ptr = malloc( n_tup * sizeof(Experimental) );
    if( !exper_ptr )
        return NULL;

    for( h = 0; h < n_tup; h++ ) {
      i = 0;
      while( 0 != strcmp(exper_items[i], "" ) ) {
        for( j = 0; j < n_field; j++ ) {
         if( 0 == strcmp(exper_items[i], tup[h].item_name[j]) ) {
            switch( i ) {
              case 1:
                 scan_number = (int)strtol(tup[h].value[j],
                                                 (char **)NULL, 10);
                 for( k = NONE; k < n_scans; k++ ) {
                      if( k == scan_number ) {
                          memset( exper_ptr[h].scan_mode, '\0', SCAN_MODE_LEN);
                          strcpy( exper_ptr[h].scan_mode, scans[k].scan_name );
                      }
                 }
                 break;
              case 2:
                 vendor_number = (int)strtol(tup[h].value[j],
                                                 (char **)NULL, 10);
                 for( k = LOCAL; k < n_vend; k++ ) {
                      if( k == vendor_number ) {
                          memset( exper_ptr[h].vendor_id, '\0', VENDOR_LEN );
                          strcpy( exper_ptr[h].vendor_id, vend[k].name );
                      }
                 }
                 break;
              case 3:
                 model_number = (int)strtol(tup[h].value[j],
                                                 (char **)NULL, 10);
                 for( k = HOME_MADE; k < n_instr; k++ ) {
                      if( k == model_number ) {
                          memset( exper_ptr[h].model_id, '\0', MODEL_LEN );
                          strcpy( exper_ptr[h].model_id, instr[k].model_name );
                      }
                 }
                 break;
              case 4:
                    exper_ptr[h].temperature = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 5:
                 exper_ptr[h].n_refl_unit_cell = (short)strtol(tup[h].value[j],
                                                               (char **)NULL, 10);
                 break;
              case 6:
                    exper_ptr[h].theta_min_unit_cell = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 7:
                    exper_ptr[h].theta_max_unit_cell = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 8:
                    exper_ptr[h].crystal_size_max = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 9:
                    exper_ptr[h].crystal_size_med = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 10:
                    exper_ptr[h].crystal_size_min = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 11:
                    exper_ptr[h].crystal_density = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 12:
                    exper_ptr[h].F000 = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 13:
                    exper_ptr[h].mu = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 14:
                    exper_ptr[h].transmission_min = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 15:
                    exper_ptr[h].transmission_max = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 16:
                 exper_ptr[h].reflns_measured = (int)strtol(tup[h].value[j],
                                                               (char **)NULL, 10);
                 break;
              case 17:
                 exper_ptr[h].reflns_unique = (int)strtol(tup[h].value[j],
                                                          (char **)NULL, 10);
                 break;
              case 18:
                 exper_ptr[h].reflns_observed = (int)strtol(tup[h].value[j],
                                                            (char **)NULL, 10);
                 break;
              case 19:
                    exper_ptr[h].sigma_cutoff = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 20:
                    exper_ptr[h].theta_min = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 21:
                    exper_ptr[h].theta_max = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 22:
                    exper_ptr[h].r_merge = strtod( tup[h].value[j], (char **)NULL );
                 break;
            }
         }
        }
        i++;
      }
    }
    return exper_ptr;
}

Refinement *ProcessRefinementTuple( Tuple *tup, int n_tup, int n_field )
{
    int h = 0,
        i = 1,
        j = 0;

    Refinement *refine_ptr;

    static char *refine_items[] = { "unit_cell_seq", "refine_type", "matrix_type",
                                    "weight_scheme", "weight_coeff", "refine_reflns_incl",
                                    "n_parameters", "r_conventional", "r_weighted", "gof",
                                    "shift_err", "final_df_max", "final_df_min", "ext_coeff",
                                    "sigma_ext", "chiral_coeff", "sigma_chiral", "cif_file",
                                    "crt_file", "" };

    refine_ptr = malloc( n_tup * sizeof(Refinement) );
    if( !refine_ptr )
        return NULL;

    for( h = 0; h < n_tup; h++ ) {
      i = 0;
      while( 0 != strcmp(refine_items[i], "") ) {
        for( j = 0; j < n_field; j++ ) {
         if( 0 == strcmp(refine_items[i], tup[h].item_name[j]) ) {
            switch( i ) {
              case 1:
                 memset( refine_ptr[h].refine_type, '\0', REFINE_LEN );
                 strcpy( refine_ptr[h].refine_type, tup[h].value[j] );
                 break;
              case 2:
                 memset( refine_ptr[h].matrix_type, '\0', MATRIX_LEN );
                 strcpy( refine_ptr[h].matrix_type, tup[h].value[j] );
                 break;
              case 3:
                 memset( refine_ptr[h].weight_scheme, '\0',WEIGHT_LEN );
                 strcpy( refine_ptr[h].weight_scheme, tup[h].value[j] );
                 break;
              case 4:
                 refine_ptr[h].weight_coeff = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 5:
                 refine_ptr[h].refine_reflns_incl = (int)strtol(tup[h].value[j],
                                                               (char **)NULL, 10);
                 break;
              case 6:
                 refine_ptr[h].n_parameters = (int)strtol(tup[h].value[j],
                                                               (char **)NULL, 10);
                 break;
              case 7:
                 refine_ptr[h].r_conventional = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 8:
                 refine_ptr[h].r_weighted = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 9:
                 refine_ptr[h].gof = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 10:
                 refine_ptr[h].shift_err = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 11:
                 refine_ptr[h].final_df_max = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 12:
                 refine_ptr[h].final_df_min = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 13:
                 refine_ptr[h].ext_coeff = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 14:
                 refine_ptr[h].sigma_ext = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 15:
                 refine_ptr[h].chiral_coeff = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 16:
                 refine_ptr[h].sigma_chiral = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 17:
                 memset( refine_ptr[h].cif_file, '\0', CIF_FILE_LEN );
                 strcpy( refine_ptr[h].cif_file, tup[h].value[j] );
                 break;
              case 18:
                 memset( refine_ptr[h].crt_file, '\0', CRT_FILE_LEN );
                 strcpy( refine_ptr[h].crt_file, tup[h].value[j] );
                 break;
            }
         }
        }
        i++;
      } 
    }
    return refine_ptr;
}

AtomCoords *ProcessAtomCoordsTuple(Tuple *tup, int n_tup, int n_field )
{
    int h = 0,
        i = 1,
        j = 0;

    AtomCoords *atom_ptr;
    static char *atom_coords_items[] =
    { "unit_cell_seq", "atomic_number", "atom_name", "site_symmetry",
      "fix_u23", "fix_u13", "fix_u12", "fix_u33",
      "fix_u22", "fix_u11", "fix_z", "fix_y", "fix_x", "fix_occ",
      "adp_flag", "occupancy", "x", "y", "z", "sigma_occupancy",
      "sigma_x", "sigma_y", "sigma_z", "u11", "u22", "u33",
      "u12", "u13", "u23", "sigma_u11", "sigma_u22", "sigma_u33",
      "sigma_u12", "sigma_u13", "sigma_u23",
      "" };
    
    atom_ptr = malloc( n_tup * sizeof(AtomCoords) );
    if( !atom_ptr )
        return NULL;

    for( h = 0; h < n_tup; h++ ) {
      i = 0;
      atom_ptr[h].refine_flags = 0;   
      while( 0 != strcmp(atom_coords_items[i], "") ) {
        for( j = 0; j < n_field; j++ ) {
         if( 0 == strcmp(atom_coords_items[i], tup[h].item_name[j]) ) {
            switch( i ) {
              case 1:
                 atom_ptr[h].atomic_number = (short)strtol(tup[h].value[j],
                                                          (char **)NULL, 10);
                 break;
              case 2:
                 memset( atom_ptr[h].atom_name, '\0', ATOM_NAME_LEN );
                 strcpy( atom_ptr[h].atom_name, tup[h].value[j] );
                 break;
              case 3:
                 memset( atom_ptr[h].site_symmetry, '\0', SITE_SYMM_LEN );
                 strcpy( atom_ptr[h].site_symmetry, tup[h].value[j] );
                 break;
              case 4:
                 if( 'y' == tup[h].value[j][0] ) 
                    atom_ptr[h].refine_flags |= ( 1 << 0 );
                 break;
              case 5:
                 if( 'y' == tup[h].value[j][0] ) 
                    atom_ptr[h].refine_flags |= ( 1 << 1 );
                 break;
              case 6:
                 if( 'y' == tup[h].value[j][0] ) 
                    atom_ptr[h].refine_flags |= ( 1 << 2 );
                 break;
              case 7:
                 if( 'y' == tup[h].value[j][0] ) 
                    atom_ptr[h].refine_flags |= ( 1 << 3 );
                 break;
              case 8:
                 if( 'y' == tup[h].value[j][0] ) 
                    atom_ptr[h].refine_flags |= ( 1 << 4 );
                 break;
              case 9:
                 if( 'y' == tup[h].value[j][0] ) 
                    atom_ptr[h].refine_flags |= ( 1 << 5 );
                 break;
              case 10:
                 if( 'y' == tup[h].value[j][0] ) 
                    atom_ptr[h].refine_flags |= ( 1 << 6 );
                 break;
              case 11:
                 if( 'y' == tup[h].value[j][0] ) 
                    atom_ptr[h].refine_flags |= ( 1 << 7 );
                 break;
              case 12:
                 if( 'y' == tup[h].value[j][0] ) 
                    atom_ptr[h].refine_flags |= ( 1 << 8 );
                 break;
              case 13:
                 if( 'y' == tup[h].value[j][0] ) 
                    atom_ptr[h].refine_flags |= ( 1 << 9 );
                 break;
              case 14:
                 if( 'a' == tup[h].value[j][0] ) 
                    atom_ptr[h].refine_flags |= ( 1 << 10 );
                 break;
              case 15:
                 atom_ptr[h].occupancy = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 16:
                 atom_ptr[h].x = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 17:
                 atom_ptr[h].y = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 18:
                 atom_ptr[h].z = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 19:
                 atom_ptr[h].sigma_occupancy = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 20:
                 atom_ptr[h].sigma_x = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 21:
                 atom_ptr[h].sigma_y = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 22:
                 atom_ptr[h].sigma_z = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 23:
                 atom_ptr[h].u11 = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 24:
                 atom_ptr[h].u22 = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 25:
                 atom_ptr[h].u33 = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 26:
                 atom_ptr[h].u12 = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 27:
                 atom_ptr[h].u13 = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 28:
                 atom_ptr[h].u23 = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 29:
                 atom_ptr[h].sigma_u11 = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 30:
                 atom_ptr[h].sigma_u22 = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 31:
                 atom_ptr[h].sigma_u33 = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 32:
                 atom_ptr[h].sigma_u12 = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 33:
                 atom_ptr[h].sigma_u13 = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 34:
                 atom_ptr[h].sigma_u23 = strtod( tup[h].value[j], (char **)NULL );
                 break;
            }
         }
        }
         i++;
      }
    }
    return atom_ptr;
}

BondLength *ProcessBondLengthTuple(Tuple *tup, int n_tup, int n_field )
{
    int h = 0,
        i = 1,
        j = 0;

    BondLength  *bond_ptr;

    static char *bond_length_items[] = 
                { "unit_cell_seq", "atom1", "atom2", "distance", "su", "" };

    bond_ptr = malloc( n_tup * sizeof(BondLength) );
    if( !bond_ptr) 
       return NULL;
 
    for( h = 0; h < n_tup; h++ ) {
      i = 0;
      while( 0 != strcmp(bond_length_items[i], "") ) {
        for( j = 0; j < n_field; j++ ) {
         if( 0 == strcmp(bond_length_items[i], tup[h].item_name[j]) ) {
            switch( i ) {
              case 1:
                 memset( bond_ptr[h].atom_1, '\0', ATOM_NAME_LEN );
                 strcpy( bond_ptr[h].atom_1, tup[h].value[j] );
                 break;
              case 2:
                 memset( bond_ptr[h].atom_2, '\0', ATOM_NAME_LEN );
                 strcpy( bond_ptr[h].atom_2, tup[h].value[j] );
                 break;
              case 3:
                 bond_ptr[h].dist = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 4:
                 bond_ptr[h].esd = strtod( tup[h].value[j], (char **)NULL );
                 break;
            }
         }
        }
         i++;
      }
    }
    return bond_ptr;
}

BondAngle *ProcessBondAngleTuple(Tuple *tup, int n_tup, int n_field )
{
    int h = 0,
        i = 1,
        j = 0;

    BondAngle *angle_ptr;
         
    static char *bond_angle_items[] =
                { "unit_cell_seq", "atom1", "atom2", "atom3", "angle", "su", ""  };

    angle_ptr = malloc( n_tup * sizeof(BondAngle) );
    if( !angle_ptr )
        return NULL;

    for( h = 0; h < n_tup; h++ ) {
      i = 0;
      while( 0 != strcmp(bond_angle_items[i], "") ) {
        for( j = 0; j < n_field; j++ ) {
         if( 0 == strcmp(bond_angle_items[i], tup[h].item_name[j]) ) {
            switch( i ) {
              case 1:
                 memset( angle_ptr[h].atom_1, '\0', ATOM_NAME_LEN );
                 strcpy( angle_ptr[h].atom_1, tup[h].value[j] );
                 break;
              case 2:
                 memset( angle_ptr[h].atom_2, '\0', ATOM_NAME_LEN );
                 strcpy( angle_ptr[h].atom_2, tup[h].value[j] );
                 break;
              case 3:
                 memset( angle_ptr[h].atom_3, '\0', ATOM_NAME_LEN );
                 strcpy( angle_ptr[h].atom_3, tup[h].value[j] );
                 break;
              case 4:
                 angle_ptr[h].angle = strtod( tup[h].value[j], (char **)NULL );
                 break;
              case 5:
                 angle_ptr[h].esd = strtod( tup[h].value[j], (char **)NULL );
                 break;
            }
         }
        }
        i++;
      }
    }
    return angle_ptr;
}
/* FreeTuple -- frees all the allocated memory associated with
 * a Tuple structure.
 */

void FreeTuple( Tuple *tup, int nt, int nf )
{
    int i,
        j;
        

    for( i = nt - 1; i >= 0; i-- ) {
       for( j = nf - 1; j >= 0; j-- ) {
            if( tup[i].item_name[j] )
                free( tup[i].item_name[j] );
            if( tup[i].value[j] )
                free( tup[i].item_name[j] );
       }
       if( tup[i].item_name )
           free( tup[i].item_name );
       if( tup[i].value )
           free( tup[i].value );
    }
    if( tup )
        free( tup );

    return;
}

