/* SearchDatabase searches the database using the libpq
 * interface.   
 * Written by:
 *	Paul D. Boyle
 *	Department of Chemistry
 *	Box 8204
 *	North Carolina State University
 *	Raleigh, NC 27695-8204
 *
 *	March 1998
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#define _XOPEN_SOURCE

#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>


/* postgresql header */
#include "/usr/local/pgsql/include/libpq-fe.h"

/* local headers */
#include "sinfo.h"
#include "cgi_util.h"
#include "qthread.h"

/* globals */
extern int run_mode;

Tuple *SearchDatabase( Query *query,
                       int *n_tuples, int *n_fields )
{

    int              j, k;

    size_t           len;

    char             *pg_tty = NULL,
                     *pg_port = NULL,
                     *pg_options = NULL,
                     *tmp;

    Tuple            *tuples = NULL;  /* to collect the tuple found */

/* some libpq specific variables */
                 
    PGconn           *conn;
    ConnStatusType   pg_status;
    PGresult         *pg_res;

/* open the connection to the database */
    conn = PQsetdb( query->db_host, pg_port, pg_options,
                    pg_tty, query->db_name);

    pg_status = PQstatus( conn );
    if( CONNECTION_BAD == pg_status ) {
#ifdef DEBUG
        if( CLI_MODE == run_mode ) 
            fputs( "Bad Connection to Database\n", stderr );
        else
            ReportErrorHTML( "Bad Connection to Database" );
#endif
        return NULL;
    }
/* now do the query to the database */

    pg_res = PQexec( conn, query->query_buffer );


    if( PGRES_TUPLES_OK != PQresultStatus(pg_res) )  {
#ifdef DEBUG
        if( CLI_MODE == run_mode ) 
            fputs( "Database Query was Unsuccessful\n", stderr );
        else
           ReportErrorHTML( "Database Query was Unsuccessful" );
#endif
        PQclear( pg_res );
        PQfinish( conn ); 
        return NULL;
    }

/* interrogate the pg_res struct and copy the results to a Tuple
 * structure.
 */

    *n_tuples = PQntuples( pg_res );
    *n_fields = PQnfields( pg_res );

    if( 0 == *n_tuples ) return NULL;

    tuples = malloc(*n_tuples * sizeof(Tuple) );
    if( !tuples ) 
       return NULL;


    for( j = 0; j < *n_tuples; j++ ) { 
       tuples[j].item_name = malloc( *n_fields * sizeof( tuples->item_name ) );
       if( !tuples[j].item_name ) 
          return NULL;
       tuples[j].value = malloc( *n_fields * sizeof( tuples->value ) );
       if( !tuples[j].value ) 
          return NULL;

       for( k = 0; k < *n_fields; k++ ) {
          tmp = PQfname( pg_res, k );
          len = strlen( tmp );
          tuples[j].item_name[k] = malloc( len + 1);
          if( !tuples[j].item_name[k] ) 
              return NULL;
          strcpy( tuples[j].item_name[k], tmp );

          tmp = PQgetvalue( pg_res, j, k );
          len = strlen( tmp );
          tuples[j].value[k] = malloc( len + 1);
          if( !tuples[j].value[k] ) 
              return NULL;
          strcpy( tuples[j].value[k], tmp );
       }
    }         

/* cleanup PostgreSQL structures and other stuff */
    PQclear( pg_res );
    PQfinish( conn );

    return tuples;   
}

/* this function takes a structure identification code as an
 * argument and returns the unit_cell_seq number.
 * A -1 is returned if an error occured.  This function
 * contains PostgreSQL specific code.
 */

int RetrieveUnitCellSeq( char *ident )
{
    int unit_cell_seq;

/* some libpq specific variables */

    PGconn           *conn;
    ConnStatusType   pg_status;
    PGresult         *pg_res;

/* open the connection to the database */
    conn = PQsetdb( query->db_host, pg_port, pg_options,
                    pg_tty, query->db_name);

    pg_status = PQstatus( conn );
    if( CONNECTION_BAD == pg_status ) {
#ifdef DEBUG
        if( CLI_MODE == run_mode )
            fputs( "Bad Connection to Database\n", stderr );
        else
            ReportErrorHTML( "Bad Connection to Database" );
#endif
        return -1;
    }
/* now do the query to the database */

    pg_res = PQexec( conn, query->query_buffer );


    if( PGRES_TUPLES_OK != PQresultStatus(pg_res) )  {
#ifdef DEBUG
        if( CLI_MODE == run_mode )
            fputs( "Database Query was Unsuccessful\n", stderr );
        else
           ReportErrorHTML( "Database Query was Unsuccessful" );
#endif
        PQclear( pg_res );
        PQfinish( conn );
        return -1;
    }

}

