/* WholeTableThread -- is a function which does a whole table
 * scan on a given table then brings the data into the appropriate
 * crystallographic data struct as defined in sinfo.h.
 * Written by:
 *	Paul D. Boyle
 *	Department of Chemistry
 *	Box 8204
 *	North Carolina State University
 *	Raleigh, NC 27695-8204
 *
 *	March 1998
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#define _XOPEN_SOURCE

#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/ipc.h>
#include <sys/msg.h>

#include "sinfo.h"
#include "qthread.h"
#include "cgi_util.h"
#include "process_tuples.h"
#include "output.h"
#include "sinfo_ipc.h"

#define SINGLE_QUOTE 0x27

/* globals */
extern int run_mode;

void WholeTableThread( void *data)
{
     Query *query = (Query *)data;
     Tuple *tuples;

/* Pointers to the various structs that contain 
 * crystallographic data.
 */
     Lattice      *latt;
     UnitCell     unitcell_data,
                  *unit_cell;
     Formula      *form;
     Experimental *exper;
     Refinement   *refine;
     AtomCoords   *atoms;
     BondLength   *bonds;
     BondAngle    *angles;

/* structures for the messages */

struct latt_msg  msg_data_latt;
struct cell_msg  msg_data_cell;
struct form_msg  msg_data_form;
struct expr_msg  msg_data_expr;
struct refn_msg  msg_data_refn;
     
  
     char table_searched[DB_STD2_LEN]; 
     char *structure_name,
          *ptr_start,
          *ptr_end;
#ifdef DEBUG
     char error_msg[80];
#endif
     char *err_ptr;
     char *table_names[] = { "lattice", "unit_cell",  "formula",
                            "experimental", "refinement", "atom_coords",
                            "bond_length", "bond_angle"};

     int n_tuples, 
         n_fields,
         ret_val,
         msg_id,
         i;
    
     void    *recvd_buffer;  

     size_t len;

/* Unpack the message queue identifier, and get the structure ident
 * code.
 */
     msg_id = query->msg_id;

     ptr_start = strchr( query->query_buffer, SINGLE_QUOTE );
     ptr_start++;
     ptr_end = strrchr( query->query_buffer, SINGLE_QUOTE );
     len = ptr_end - ptr_start;
     structure_name = malloc( len + 1 );
     strncpy( structure_name, ptr_start, len );
     structure_name[len] = '\0';


     for( i = LATT; i < N_TABLES; i++ ) {
        if( query->query_type == i )
            strcpy( table_searched, table_names[i] );
     }

/* Connect to the database and return the structure of tuples */     
     tuples = SearchDatabase( query, &n_tuples, &n_fields  );
     if( NULL == tuples ) {
#ifdef DEBUG
         if( CLI_MODE == run_mode ) 
             fprintf( stderr, "Search of %s table failed. Exiting.\n",
                                 table_searched );
         else {
             sprintf( error_msg, "Search of %s table failed.\n",
                                 table_searched );
             ReportErrorHTML( error_msg );
         }
#endif
     }

/* Now get the data stuffed in tuple out as a C struct of the proper
 * type depending on the query_type.
 */

    switch( query->query_type ) {
      case LATT:
         if( NULL == tuples ) {  /* notify parent of error */
             msg_data_latt.type = (long)ERR_MSG;
             err_ptr = "lattice search is NULL";
             msg_data_latt.n_bytes = strlen( err_ptr );
             memset( &msg_data_latt.latt_data, '\0', msg_data_latt.n_bytes + 1 );
             memcpy( (char *)&msg_data_latt.latt_data, err_ptr, msg_data_latt.n_bytes );
             msgsnd( msg_id, &msg_data_latt, sizeof(Lattice) +
                             sizeof(msg_data_latt.n_bytes), 0 );
             _exit( EXIT_FAILURE );
         }
         latt = ProcessLatticeTuple( tuples, n_tuples, n_fields );
         if( !latt) 
           exit( LATT_MSG );

         msg_data_latt.type = (long)LATT_MSG;
         msg_data_latt.n_bytes = sizeof( Lattice );
         msg_data_latt.latt_data = *latt;
         msgsnd( msg_id, &msg_data_latt, sizeof(Lattice) +
                         sizeof(msg_data_latt.n_bytes), 0 );
         break;
      case UNIT_CELL:
         if( NULL == tuples ) {  /* notify parent of error */
             msg_data_cell.type = (long)ERR_MSG;
             err_ptr = "cell search is NULL";
             msg_data_cell.n_bytes = strlen( err_ptr );
             memset( &msg_data_cell.cell_data, '\0', msg_data_cell.n_bytes + 1 );
             memcpy( (char *)&msg_data_cell.cell_data, err_ptr, msg_data_cell.n_bytes );
             msgsnd( msg_id, &msg_data_cell, sizeof(UnitCell) +
                             sizeof(msg_data_cell.n_bytes), 0 );
             _exit( CELL_MSG );
         }
         unit_cell = ProcessUnitCellTuple( tuples, n_tuples, n_fields );
         if( !unit_cell ) 
           exit( EXIT_FAILURE );

         msg_data_cell.type = (long)CELL_MSG;
         msg_data_cell.n_bytes = sizeof( UnitCell );
         msg_data_cell.cell_data = *unit_cell;
         msgsnd( msg_id, &msg_data_cell, sizeof(UnitCell) +
                         sizeof(msg_data_cell.n_bytes), 0 );

         /* also send data to atomic coordinates job */
         msg_data_cell.type = (long)SPEC_MSG;
         msgsnd( msg_id, &msg_data_cell, sizeof(UnitCell) +
                         sizeof(msg_data_cell.n_bytes), 0 );
         break;
      case FORM: 
         if( NULL == tuples ) {  /* notify parent of error */
             msg_data_form.type = (long)ERR_MSG;
             err_ptr = "formula search is NULL";
             msg_data_form.n_bytes = strlen( err_ptr );
             memset( msg_data_form.formula_data, '\0', msg_data_form.n_bytes + 1 );
             memcpy( (char *)msg_data_form.formula_data, err_ptr, msg_data_form.n_bytes );
             msgsnd( msg_id, &msg_data_form, sizeof(Formula) +
                             sizeof(msg_data_form.n_bytes), 0 );
             _exit( FORM_MSG );
         }
         form = ProcessFormulaTuple( tuples, n_tuples, n_fields );
         if( !form ) 
           exit( EXIT_FAILURE );

         if( n_tuples > N_ATOM_TYPES_MAX ) /* silently truncate the excess atom types */
                      n_tuples = N_ATOM_TYPES_MAX;

         msg_data_form.type = (long)FORM_MSG;
         msg_data_form.n_bytes = n_tuples * sizeof( Formula );
         memcpy(msg_data_form.formula_data, form, (unsigned)msg_data_form.n_bytes );
         msgsnd( msg_id, &msg_data_form, msg_data_form.n_bytes +
                         sizeof(msg_data_form.n_bytes), 0 );
         break;
      case EXPER:
         if( NULL == tuples ) {  /* notify parent of error */
             msg_data_expr.type = (long)ERR_MSG;
             err_ptr = "expr search is NULL";
             msg_data_expr.n_bytes = strlen( err_ptr );
             memset( &msg_data_expr.exper_data, '\0', msg_data_expr.n_bytes + 1 );
             memcpy( (char *)&msg_data_expr.exper_data, err_ptr, msg_data_expr.n_bytes );
             msgsnd( msg_id, &msg_data_expr, sizeof(Experimental) +
                             sizeof(msg_data_expr.n_bytes), 0 );
             _exit( EXPR_MSG );
         }
         exper = ProcessExperimentalTuple( tuples, n_tuples, n_fields );
         if( !exper ) 
           exit( EXIT_FAILURE );

         msg_data_expr.type = (long)EXPR_MSG;
         msg_data_expr.n_bytes = sizeof( Experimental );
         msg_data_expr.exper_data = *exper;
         msgsnd( msg_id, &msg_data_expr, sizeof(Experimental) +
                         sizeof(msg_data_expr.n_bytes), 0 );
         break;
      case REFINE:
         if( NULL == tuples ) {  /* notify parent of error */
             msg_data_refn.type = (long)ERR_MSG;
             err_ptr = "refine search is NULL";
             msg_data_refn.n_bytes = strlen( err_ptr );
             memset( &msg_data_refn.refine_data, '\0', msg_data_refn.n_bytes + 1 );
             memcpy( (char *)&msg_data_refn.refine_data, err_ptr, msg_data_refn.n_bytes );
             msgsnd( msg_id, &msg_data_refn, sizeof(Refinement) +
                             sizeof(msg_data_refn.n_bytes), 0 );
             _exit( REFN_MSG );
         }
         refine = ProcessRefinementTuple( tuples, n_tuples, n_fields );
         if( !refine ) 
           exit( EXIT_FAILURE );

         msg_data_refn.type = (long)REFN_MSG;
         msg_data_refn.n_bytes = sizeof( Refinement );
         msg_data_refn.refine_data = *refine;
         msgsnd( msg_id, &msg_data_refn, sizeof(Refinement) +
                         sizeof(msg_data_refn.n_bytes), 0 );
         break;
      case ATOM_COORDS:
         if( NULL == tuples ) 
             exit( EXIT_FAILURE );
         atoms = ProcessAtomCoordsTuple( tuples, n_tuples, n_fields );
         if( !atoms ) 
           exit( EXIT_FAILURE );

         recvd_buffer = malloc( QUEUE_SIZ );
         if( !recvd_buffer ) {
             if( CLI_MODE == run_mode )
                 fprintf( stderr, "Memory Allocation Error, %s:%d\n",
                                  __FILE__, __LINE__ );
             else
                 ReportErrorHTML( "Memory Allocation Error.\n" );
         }

         /* get data from cell search for diagonalizing the Uij's */
         ret_val = msgrcv( msg_id, recvd_buffer, QUEUE_SIZ, SPEC_MSG, 0 );
         msg_data_cell = *(struct cell_msg *)recvd_buffer;
         unitcell_data = msg_data_cell.cell_data;

         ProduceAtomParamsFile( structure_name, atoms, &unitcell_data, n_tuples );
         break;
      case BOND_LENGTH:
         if( NULL == tuples ) 
             exit( EXIT_FAILURE );
         bonds = ProcessBondLengthTuple( tuples, n_tuples, n_fields );
         if( ! bonds ) 
           exit( EXIT_FAILURE );

         MolecularGeometryTable( structure_name, bonds, n_tuples, BOND_LENGTH );
         break;
      case BOND_ANGLE:
         if( NULL == tuples ) 
             exit( EXIT_FAILURE );
         angles = ProcessBondAngleTuple( tuples, n_tuples, n_fields );
         if( ! angles ) 
           exit( EXIT_FAILURE );

         MolecularGeometryTable( structure_name, angles, n_tuples, BOND_ANGLE );
         break;
      default:
         break;  /* shouldn't get here */
    }
    exit( EXIT_SUCCESS );
}
