/* Some utility functions to process CGI input from Web forms.  Based
 * in part on ideas from Thomas Boutell's book CGI Programming in 
 * C & Perl pages 80-88.
 * Written by:
 *	Paul D. Boyle
 *	Department of Chemistry
 *	Box 8204
 *	North Carolina State University
 *	Raleigh, NC 27695-8204
 *
 *	July 1997
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#define _XOPEN_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* local includes */
#include "cgi_util.h"

/* VerifyForm verifies that REQUEST_METHOD and CONTENT_TYPE environment
 * variables are properly set. Lets caller deal with error handling. Caller
 * needs to specify "GET" or "POST" as the "method"
 * Error codes: 0 == OK; 1 == wrong request method; 2 == not encoded; 
 * 3 == both method and encoding are wrong.
 */

int VerifyForm( const char *method )
{
    int err_code = 0;
    
    if( 0 != strcmp( method, getenv("REQUEST_METHOD") ) )
        err_code |= 1;

    if( 0 != strcmp( ENCODED, getenv("CONTENT_TYPE") ) )
        err_code |= 2;

    return err_code;
} 

/* Takes care of url encoded spaces. */
void PlusToSpace( char *a )
{
    while( NULL != ( a = strchr( a, SPACE_TOKEN ) ) )  *a++ = ' ';
    return;
}

/* Returns the number of fields delimited by FIELD_TOKEN there are
 * in an input buffer.  This is for dynamic memory allocation of an
 * array cgi_entry structs.  Note that cgi encoded strings are not
 * '&' delimited for the last field.  That is, the '\0' for the 
 * string tells when the input string is finished.
 */
int CountFields( char *a )
{
    int n_fields = 0;
    char *ptr;

    ptr = a;
    while( NULL != (ptr = strchr( ptr, FIELD_TOKEN )) ) {
         ptr++;
         n_fields++;
    }
    return( ++n_fields ); /* extra increment to account for last field. */
}

/* A few utility routines for writing standard HTML related stuff to 
 * a Web Page.
 */
void StdContentHeader( const char *type )
{
    fprintf( stdout, "Content-type: %s/html\n\n", type );
}

void StdHTMLHeader( void )
{
    fprintf( stdout, "<html>\n<head><title>CGI Output\n"
                     "</title></head>\n<body>" );
}

void StdHTMLFooter( void )
{
    fprintf( stdout, "</body></html>\n" );
    fflush( stdout );
}

/*Reports caller specified error messages. */
void ReportErrorHTML( const char *a ) 
{

    StdContentHeader( "text" );
    StdHTMLHeader( );
    fprintf( stdout, "<H1>Error</H1>%s\n", a );
    StdHTMLFooter();
}

/* Takes any %xx string and unescapes to it's hex ascii equivalent char
 * which is then inserted into the destination buffer.  Error returns 
 * NULL; let caller handle the error.
 */

char *TranslateEscapes( const char *src )
{
     char number[3];
     char *ret_ptr;
     char *dest = malloc( strlen(src) + 1 );

     if( !dest ) return NULL;

     ret_ptr = dest;

     while( *src ) {
       if( ESCAPE_SEQ == *src ) { /* there are characters to be unescaped */
         src++;
         strncpy( number, src, SEQ_SIZE );
         number[2] = '\0';
         *dest = (char)strtol( number, (char **)NULL, 16 );
         dest++; 
         src += SEQ_SIZE;
       }
       else 
          *dest++ = *src++;
     }
     return ret_ptr;
}

/* TranslateNewlines() takes care of differences in the end of line 
 * characters for differing platforms, DOS = CR\LF; MAC = CR; UNIX = LF
 * This function should be called *only after* the escapes have been 
 * translated.  
 */

char *TranslateNewlines( const char *src )
{
    char *ret_ptr;
    char *dest = malloc( strlen(src) + 1 );
    
    if( !dest ) return NULL; 

    ret_ptr = dest;

    while( *src ) {
       if( CARRIAGE_RET == *src ) {         /* for Mac's & DOS */
           *dest++ = '\n';     /* Use local representation for newline */
           src++;
           if( LINE_FEED == *src )  src++;  /* skip over 2nd DOS byte */
       }
       else 
         *dest++ = *src++;
    }
    return ret_ptr;
}

/* ParseNameValuePairs() breaks up the name value pairs separated by
 * the PAIR_TOKEN (i.e. '=').  This assumes the string is the union 
 * field of a cgi_data union.  No new memory needs to be allocated.
 * This assumes that the original url encoded
 * string has already been parsed into separate 'name=value' pairs. 
 */

char *ParseNameValuePairs( char *a )
{
    char *ptr;
    ptr = strchr( a, PAIR_TOKEN );
    if( ptr != NULL ) {
      *ptr = '\0';
      return ++ptr;
    }
    else
      return "";
}

/* BufferOverFlowError -- used for trapping buffer overruns
 */
void BufferOverFlowError( union cgi_data *data, int n_field )
{
    FILE *mail;
    int i;

/* Changes these pointers as is appropriate for your local situation. */

    char *from_who = FROM_WHO;
    char *from_who_email = FROM_WHO_EMAIL;
    char *recipient = RECIPIENT;

    mail = popen( SENDMAIL, "w" );
    if( !mail )
        exit( EXIT_FAILURE );

    ReportErrorHTML( "Inappropriate Input. Exiting." );

/* Write sendmail headers */
       fprintf( mail, "From: %s <%s>\nTo: %s\n"
             "Subject: Inappropriate Input Error\n\n", from_who, from_who_email,
              recipient );

/* get information about user giving inappropriate input */

    fprintf( mail,
           "\nServer Protocol: %s\n", getenv("SERVER_PROTOCOL") );
    fprintf( mail,
            "Server Port: %s\n", getenv("SERVER_PORT") );
    fprintf( mail,
            "Remote Host: %s\n", getenv("REMOTE_HOST") );
    fprintf( mail,
            "Remote IP Address: %s\n", getenv("REMOTE_ADDR") );

/* Dump the contents of the data buffer to see what was sent to the form */
    for( i = 0; i < n_field; i++ )
         fprintf( mail, "%s = %s\n",
                  data[i].pair.name, data[i].pair.value );

    pclose( mail );

    exit( EXIT_SUCCESS );
}
