/* Some utility routines  to be used in the sinfo CGI program.
 * Written by:
 *	Paul D. Boyle
 *	Department of Chemistry
 *	Box 8204
 *	North Carolina State University
 *	Raleigh, NC 27695-8204
 *
 *	March 1998
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#define _XOPEN_SOURCE

#include <sys/types.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/wait.h>


#ifndef HAVE_SINFO_H
#define HAVE_SINFO_H 1
#include "sinfo.h"
#endif

/* some useful math related defines. */
#ifndef PI
#define PI 3.1415926536
#endif

/* configure this at compile time */
#ifdef USE_B_TERMS
#define EIGHT_PI_SQ  (8.0 * PI * PI)  /* Convert U(iso) to B(iso) */
#else 
#define EIGHT_PI_SQ (1.0);
#endif

#define DEG2RAD(deg) ( (PI/180.) * (deg) )
#define RAD2DEG(rad) ( (180./PI) * (rad) )


/* see the c.l.c FAQ question 14.5 for floating point comparisions, this 
 * where these defines came from as well as the RelativeDifference() 
 * function (see below).
 */
#define ABS(x)     ((x < 0 ? -(x) : (x) ))
#define MAX(a, b)  ((a) > (b) ? (a) : (b))

#define IUCR_THRESH  20.0  /* number of sigfigs */

/* globals */
extern int run_mode;

/* FormulaString() -- takes a formula structure and outputs a string
 * representation of the formula.  Caller needs to free string.
 */
char *FormulaString( Formula *form, int n_atom_types )
{
    int i;   
    char *ret_val,
         *tmp_buffer,
         *src, 
         *dst
;
    char elem[10];        /*this should be enough room */
    ret_val = malloc( (size_t)n_atom_types * 10 );
    tmp_buffer = malloc( (size_t)n_atom_types * 10 );
    
    if( (!ret_val) || (!tmp_buffer) )
        return NULL;

/* clean the arrays of stray garbage */
    memset( elem, '\0', 10 );
    memset( ret_val, '\0', 10 );
    memset( tmp_buffer, '\0', 10 );

/* get the raw string all into one buffer */
    for( i = 0; i < n_atom_types; i++ ) {
         sprintf( elem, "%s%5.2f", form[i].element.symbol, form[i].count );
         strcat( tmp_buffer, elem );
         strcat ( tmp_buffer, " " );
    }
    src = tmp_buffer;
    dst = ret_val;

    while( *src ) {
       if( ('.' == *src) &&  /* special case 1: get rid of .00  */
           ('0' == *(src+1)) &&
           ('0' == *(src+2)) ) src += 3;
       else if( (isalpha(*src)) &&  /* special case 2: take out extra spaces */
                (isspace(*(src+1)) ) &&
                (isdigit(*(src+2) ) ) ) {
            *dst++ = *src++;
            src++;
       }
       else
          *dst++ = *src++;
    }
    free( tmp_buffer );

/* clean up trailing stuff */
    src = ret_val;

    while( *src ) {
      if( !isprint(*src) ) *src = '\0';
      src++;
    }
     
    return ret_val;
}

double CalcFormulaWeight( Formula *form, int n_atom_types )
{
    double wt = 0.0;
    int i;
 
    for( i = 0; i < n_atom_types; i++ )
       wt += form[i].count * form[i].element.atomic_weight;

    return wt;
}


/* RelativeDifference() -- */

double RelativeDifference( double a, double b )
{
    double c = ABS(a);
    double d = ABS(b);

    d = MAX( c, d );
    return d == 0.0 ? 0.0 : ABS( a - b) / d;
}


/* PositionalRefinementFlags() --
 * This is a function decodes the positional parameter refinemennt
 * flags and returns an integer to indicate which flags are set.
 * If a bit is set, then that means the parameter was *not*
 * allowed to refine, i.e. it was 'fixed'.
 *
 * Return code:
 * 0   - no positional parameters fixed.
 * 1   - x is fixed
 * 2   - y is fixed
 * 3   - x and y are fixed
 * 4   - z is fixed
 * 5   - x and z are fixed
 * 6   - y and z are fixed
 * 7   - x y and z are fixed
 */

int PositionalRefinementFlags( unsigned long flag )
{
    int ret_val = 0;

    /* test for x, y and z in succession */
     if( (flag & (1<<8)) != 0 )
       ret_val += 1;

     if( (flag & (1<<7)) != 0 )
       ret_val += 2;

     if( (flag & (1<<6)) != 0 )
       ret_val += 4;

     return ret_val;
}



int CalcSignificantFigures( double sigma, double *new_sig ) 
{
    int sig_fig = 0;

    if( RelativeDifference(sigma, 0.0) <= TOLERANCE ) {
        *new_sig = 0.0;
         return -1;       /* parameter has no sigma, (not refined) */
    }

    while( sigma < IUCR_THRESH ) {
        sigma *= 10.0;
        sig_fig++;
    }
    *new_sig = sigma / 10.0;
    return --sig_fig;
}

/* IsAnisotropic() -- function to test whether a set of U's were
 * refined anisotropically.  Function returns a 1 if atom has been refined
 * anisotropically, 0 if it has been refined isotropically.
 */
int IsAnisotropic( unsigned long flag )
{
  return flag & (1<<10);
}

/* for now, fake Biso with simply the average of the diagonal
 * elements of the U tensor multiplied by 8Pi^2.  Implement
 * a proper principal axes calculation with beta's at a later
 * time.
 */
double CalcEquivADP( double u11, double u22, double u33, 
                     double u12, double u13, double u23 )
{
    double adp[6], 
           avg = 0.0;
        
    int i;
    
    adp[0] = u11;  /* get elements into an array for loops */
    adp[1] = u22; 
    adp[2] = u33; 
    adp[3] = u12; 
    adp[4] = u13; 
    adp[5] = u23; 

    avg = 0.0;
    for( i = 0; i < 3; i++ )
       avg += adp[i];

    return  avg / 3.0;
}

/* the expression for calculating the sigma of a cosine of an angle
 * comes from NRCVAX in the cdfile.f file.  Written presumeably by 
 * Eric Gabe.
 */
double CalcSigmaCOSAngle( double angle, double sigma_angle )
{
    double sigma_cos_angle;
    
    sigma_cos_angle = sigma_angle * sin( DEG2RAD(angle) ) * PI / 180.0;
    return sigma_cos_angle;
}

/* the expression for calculating the sigma of the unit cell volume
 * comes from NRCVAX in the report.f file.  Written presumeably by 
 * Eric Gabe.
 */
double CalcSigmaCellVolume( UnitCell *cell )
{
    double sigma_cos_alpha,
           sigma_cos_beta,
           sigma_cos_gamma,
           cos_alp, cos_bet, cos_gam,
           trig;
    double tmp_a, tmp_b, tmp_c, tmp_v;

    cell->sigma_volume = 0.0;

    cos_alp = cos( DEG2RAD(cell->alpha) );
    cos_bet = cos( DEG2RAD(cell->beta) );
    cos_gam = cos( DEG2RAD(cell->gamma) );

    trig = 1.0 - cos_alp * cos_alp - cos_bet * cos_bet - cos_gam * cos_gam
           + 2.0 * cos_alp * cos_bet * cos_gam;
  
    tmp_a = cell->b * cell->c * cell->sigma_a;
    tmp_b = cell->a * cell->c * cell->sigma_b;
    tmp_c = cell->a * cell->b * cell->sigma_c;

    tmp_v = trig * ( tmp_a * tmp_a + tmp_b * tmp_b + tmp_c * tmp_c );

    sigma_cos_alpha = CalcSigmaCOSAngle( cell->alpha,
                                         cell->sigma_alpha );

    sigma_cos_beta  = CalcSigmaCOSAngle( cell->beta,
                                         cell->sigma_beta );

    sigma_cos_gamma = CalcSigmaCOSAngle( cell->gamma,
                                         cell->sigma_gamma );

    tmp_a = (cos_bet * cos_gam - cos_alp) * sigma_cos_alpha;   
    tmp_b = (cos_alp * cos_gam - cos_bet) * sigma_cos_beta;   
    tmp_c = (cos_alp * cos_bet - cos_gam) * sigma_cos_gamma;   

    tmp_v += ( tmp_a * tmp_a + tmp_b * tmp_b + tmp_c * tmp_c );
    cell->sigma_volume = sqrt( tmp_v );

    return cell->sigma_volume;
}
