/* results.sql  -- run this file as a pgsql script to set up the
 * database for archiving and retrieving crystal structures.
 *
 * Written by:
 *	Paul D. Boyle and C. Linus Hicks
 *      Copyright 1997, 1998
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

CREATE TABLE lattice (
	unit_cell_seq	int NOT NULL,
	n_mol_cell	smallint,
	crystal_family	varchar(16),
        lattice_type	char(1),
	space_group	varchar(12) );

CREATE UNIQUE INDEX lattice_ix ON lattice
( unit_cell_seq
);

CREATE TABLE unit_cell (
	unit_cell_seq	int NOT NULL,
	id		varchar(12) NOT NULL,
	sample_date	date,
	principal	varchar(40),
	sample_orig	varchar(40),
	radiation	varchar(12) NOT NULL,
 	wavelength	float(4) NOT NULL,
	a		float(4) NOT NULL,
	b		float(4) NOT NULL,
	c		float(4) NOT NULL,
	alpha		float(4) NOT NULL,
	beta		float(4) NOT NULL,
	gamma		float(4) NOT NULL,
	volume		float(4) NOT NULL,
	reduced_vol	float(4) NOT NULL,
	sigma_a		float(4),
	sigma_b		float(4),
	sigma_c		float(4),
	sigma_alpha	float(4),
	sigma_beta	float(4),
	sigma_gamma	float(4)
        );

CREATE UNIQUE INDEX unit_cell_ix ON unit_cell
( unit_cell_seq
);

CREATE UNIQUE INDEX unit_cell_id_ix ON unit_cell
( id
);

CREATE TABLE formula (
	unit_cell_seq	int NOT NULL,
	atomic_number	smallint NOT NULL,
	count		float(4) );

CREATE UNIQUE INDEX formula_ix ON formula
( unit_cell_seq, atomic_number
);

CREATE TABLE  experimental (
	unit_cell_seq		int NOT NULL,
	scan_mode		smallint,
	vendor_id		smallint,
	model_id		smallint,
	temperature		float(4),
	n_refl_unit_cell 	smallint,
	theta_min_unit_cell	float(4),
	theta_max_unit_cell	float(4),
	crystal_size_max	float(4),
	crystal_size_med	float(4),
	crystal_size_min	float(4),
	crystal_density		float(4),
	F000			float(4),
	mu			float(4),
	transmission_min	float(4),
	transmission_max	float(4),
	reflns_measured		int,
	reflns_unique		int,
	reflns_observed		int,
	sigma_cutoff		float(4),
	theta_min		float(4),
	theta_max		float(4),
	r_merge			float(4) );

CREATE UNIQUE INDEX experimental_ix ON experimental
( unit_cell_seq
);

CREATE TABLE refinement (
	unit_cell_seq		int NOT NULL,
	refine_type		varchar(4),
	matrix_type		varchar(16),
	weight_scheme		varchar(21),
	weight_coeff		float(4),
	refine_reflns_incl	int,
	n_parameters		int,
	r_conventional		float(4),
	r_weighted		float(4),
	gof			float(4),
	shift_err		float(4),
	final_df_max		float(4),
	final_df_min		float(4),
	ext_coeff		float(4),     -- extinction coefficent
	sigma_ext		float(4),
	chiral_coeff		float(4),
	sigma_chiral		float(4),
        cif_file                varchar(80),
        crt_file                varchar(80) );

CREATE UNIQUE INDEX refinement_ix ON refinement
( unit_cell_seq
);

CREATE TABLE atom_coords (
	unit_cell_seq		int NOT NULL,
	atomic_number		smallint,
	atom_name		varchar(8),
        site_symmetry		varchar(8),
	fix_u23			varchar(1),  -- "fix_" indicates whether
	fix_u13			varchar(1),  -- parameter was fixed
	fix_u12			varchar(1),  -- these can be 'y' or 'n' 
	fix_u33			varchar(1),
	fix_u22			varchar(1),
	fix_u11			varchar(1),
	fix_z			varchar(1),
	fix_y			varchar(1),
	fix_x			varchar(1),
	fix_occ			varchar(1),
	adp_flag		varchar(1), -- can be 'a' (aniso) or
        occupancy               float(4),   -- 'i' (isotropic)
	x			float(4),
	y			float(4),
	z			float(4),
        sigma_occupancy         float(4),
	sigma_x			float(4),
	sigma_y			float(4),
	sigma_z			float(4),
	u11			float(4),
	u22			float(4),
	u33			float(4),
	u12			float(4),
	u13			float(4),
	u23			float(4),
	sigma_u11		float(4),
	sigma_u22		float(4),
	sigma_u33		float(4),
	sigma_u12		float(4),
	sigma_u13		float(4),
	sigma_u23		float(4) );

CREATE INDEX atom_coords_ix ON atom_coords
( unit_cell_seq
);

CREATE INDEX atom_coords_an_ix ON atom_coords
( atomic_number
);

CREATE TABLE bond_length (
	unit_cell_seq		int NOT NULL,
	atom1			varchar(8),
	atom2			varchar(8),
	distance		float(4),
	su			float(4) );

CREATE TABLE bond_angle (
	unit_cell_seq		int NOT NULL,
	atom1			varchar(8),
	atom2			varchar(8),
	atom3			varchar(8),
	angle			float(4),
	su			float(4) );

-- Define the "helper" tables

CREATE TABLE scan_type (
	scan_mode	int NOT NULL,
	description	varchar(24)
        );  

CREATE UNIQUE INDEX scan_type_ix ON scan_type
( scan_mode
);
 
CREATE TABLE vendor (
	vendor_id	int NOT NULL,
	name		varchar(80) 
        );

CREATE UNIQUE INDEX vendor_ix ON vendor
( vendor_id
);

CREATE TABLE instrument (
	model_id	int NOT NULL,
	name		varchar(24)
        );

CREATE UNIQUE INDEX instrument_ix ON instrument
( model_id
);

/* Now create sequences and foreign key constraints */

CREATE SEQUENCE unit_cell_sqn;

CREATE SEQUENCE scan_mode_sqn;

CREATE SEQUENCE vendor_sqn;

CREATE SEQUENCE model_sqn;

/* Temporarily disable the foreign keys 
ALTER TABLE experimental ADD (
	CONSTRAINT experimental_fk
	FOREIGN KEY ( unit_cell_seq )
	REFERENCES unit_cell ( unit_cell_seq ));

ALTER TABLE lattice ADD (
	CONSTRAINT lattice_fk
	FOREIGN KEY ( unit_cell_seq )
	REFERENCES experimental ( unit_cell_seq ));

ALTER TABLE refinement ADD (
	CONSTRAINT refinement_fk
	FOREIGN KEY ( unit_cell_seq )
	REFERENCES experimental ( unit_cell_seq ));

ALTER TABLE atomic_coords ADD (
	CONSTRAINT atomic_coords_fk
	FOREIGN KEY ( unit_cell_seq )
	REFERENCES refinement ( unit_cell_seq ));

ALTER TABLE formula ADD (
	CONSTRAINT formula_fk
	FOREIGN KEY ( unit_cell_seq )
	REFERENCES refinement ( unit_cell_seq ));

ALTER TABLE formula ADD (
	CONSTRAINT formula_an_fk
	FOREIGN KEY ( atomic_number )
	REFERENCES periodic_tab ( atomic_number ));
*/
/* Foreign constraints for the helper tables. */
/*
ALTER TABLE experimental ADD (
	CONSTRAINT experimental_sm_fk
	FOREIGN KEY ( scan_mode )
	REFERENCES scan_type ( scan_mode ));

ALTER TABLE experimental ADD (
	CONSTRAINT experimental_ven_fk
	FOREIGN KEY ( vendor_id )
	REFERENCES vendor ( vendor_id ));

ALTER TABLE experimental ADD (
	CONSTRAINT experimental_mdl_fk
	FOREIGN KEY ( model_id )
	REFERENCES instrument ( model_id ));
*/

/* Create functions to simulate the foreign key constraints. */

CREATE FUNCTION fk_exp_1(int4) RETURNS int4
	AS 'SELECT count(*) FROM unit_cell WHERE unit_cell_seq = $1'
	LANGUAGE 'sql';

CREATE FUNCTION fk_exp_2(int4) RETURNS int4
	AS 'SELECT count(*) FROM scan_type WHERE scan_mode = $1'
	LANGUAGE 'sql';

CREATE FUNCTION fk_exp_3(int4) RETURNS int4
	AS 'SELECT count(*) FROM vendor WHERE vendor_id = $1'
	LANGUAGE 'sql';

CREATE FUNCTION fk_exp_4(int4) RETURNS int4
	AS 'SELECT count(*) FROM instrument WHERE model_id = $1'
	LANGUAGE 'sql';

/* Args: unit_cell, scan_mode, vendor_id, model_id */
CREATE FUNCTION fk_experimental(int4, int4, int4, int4) RETURNS bool
	AS 'SELECT (fk_exp_1($1) + fk_exp_2($2) + fk_exp_3($3) + fk_exp_4($4)) = 4;'
	LANGUAGE 'sql';

/* Create some functions to help out inserting seed data. */

CREATE FUNCTION get_scan_mode_seq() RETURNS int4
	AS 'SELECT nextval (\'scan_mode_sqn\')' LANGUAGE 'sql';

CREATE FUNCTION get_vendor_seq() RETURNS int4
	AS 'SELECT nextval (\'vendor_sqn\')' LANGUAGE 'sql';

CREATE FUNCTION get_model_seq() RETURNS int4
	AS 'SELECT nextval (\'model_sqn\')' LANGUAGE 'sql';

/* Now populate the helper tables */

INSERT INTO scan_type
	VALUES ( get_scan_mode_seq(), 'None' );
INSERT INTO scan_type
	VALUES ( get_scan_mode_seq(), 'Theta/2Theta' );
INSERT INTO scan_type
	VALUES ( get_scan_mode_seq(), 'Omega' );
INSERT INTO scan_type
	VALUES ( get_scan_mode_seq(), 'Special' );

INSERT INTO vendor
	VALUES ( get_vendor_seq(), 'Nonius' );
INSERT INTO vendor
	VALUES ( get_vendor_seq(), 'Enraf-Nonius' );
INSERT INTO vendor
	VALUES ( get_vendor_seq(), 'Siemens' );
INSERT INTO vendor
	VALUES ( get_vendor_seq(), 'Bruker' );
INSERT INTO vendor
	VALUES ( get_vendor_seq(), 'Rigaku' );
INSERT INTO vendor
	VALUES ( get_vendor_seq(), 'Kuma' );

INSERT INTO instrument
	VALUES ( get_model_seq(), 'CAD4' );
INSERT INTO instrument
	VALUES ( get_model_seq(), 'CAD4-MACH' );
INSERT INTO instrument
	VALUES ( get_model_seq(), 'CAD4-TURBO' );
INSERT INTO instrument
	VALUES ( get_model_seq(), 'MACH3' );
INSERT INTO instrument
	VALUES ( get_model_seq(), 'KCCD' );
INSERT INTO instrument
	VALUES ( get_model_seq(), 'P4' );
INSERT INTO instrument
	VALUES ( get_model_seq(), 'P3' );
INSERT INTO instrument
	VALUES ( get_model_seq(), 'R3' );
INSERT INTO instrument
	VALUES ( get_model_seq(), 'SMART CCD' );
INSERT INTO instrument
	VALUES ( get_model_seq(), 'P21' );
INSERT INTO instrument
	VALUES ( get_model_seq(), 'KM-4' );

INSERT INTO instrument
	VALUES ( get_model_seq(), 'AFC5' );

INSERT INTO instrument
	VALUES ( get_model_seq(), 'AFC6' );

INSERT INTO instrument
	VALUES ( get_model_seq(), 'AFC7' );
